"""Test module for nested types under PyTables."""

import sys
import itertools

import numpy as np

import tables as tb
from tables.tests import common

minRowIndex = 10


# This is the structure of the table used for testing (DON'T PANIC!):
#
# +-+---------------------------------+-----+----------+-+-+
# |x|Info                             |color|info      |y|z|
# | +-----+--+----------------+----+--+     +----+-----+ | |
# | |value|y2|Info2           |name|z2|     |Name|Value| | |
# | |     |  +----+-----+--+--+    |  |     |    |     | | |
# | |     |  |name|value|y3|z3|    |  |     |    |     | | |
# +-+-----+--+----+-----+--+--+----+--+-----+----+-----+-+-+
#
# Please note that some fields are explicitly ordered while others are
# ordered alphabetically by name.
# The declaration of the nested table:
class Info(tb.IsDescription):
    _v_pos = 3
    Name = tb.StringCol(itemsize=2)
    Value = tb.ComplexCol(itemsize=16)


class TestTDescr(tb.IsDescription):
    """A description that has several nested columns."""

    x = tb.Int32Col(dflt=0, shape=2, pos=0)  # 0
    y = tb.Float64Col(dflt=1, shape=(2, 2))
    z = tb.UInt8Col(dflt=1)
    color = tb.StringCol(itemsize=2, dflt=b" ", pos=2)
    info = Info()

    class Info(tb.IsDescription):  # 1
        _v_pos = 1
        name = tb.StringCol(itemsize=2)
        value = tb.ComplexCol(itemsize=16, pos=0)  # 0
        y2 = tb.Float64Col(dflt=1, pos=1)  # 1
        z2 = tb.UInt8Col(dflt=1)

        class Info2(tb.IsDescription):
            y3 = tb.Time64Col(dflt=1, shape=2)
            z3 = tb.EnumCol({"r": 4, "g": 2, "b": 1}, "r", "int32", shape=2)
            name = tb.StringCol(itemsize=2)
            value = tb.ComplexCol(itemsize=16, shape=2)


# The corresponding nested array description:
testADescr = [
    ("x", "(2,)int32"),
    (
        "Info",
        [
            ("value", "complex128"),
            ("y2", "float64"),
            (
                "Info2",
                [
                    ("name", "S2"),
                    ("value", "(2,)complex128"),
                    ("y3", "(2,)float64"),
                    ("z3", "(2,)int32"),
                ],
            ),
            ("name", "S2"),
            ("z2", "uint8"),
        ],
    ),
    ("color", "S2"),
    ("info", [("Name", "S2"), ("Value", "complex128")]),
    ("y", "(2,2)float64"),
    ("z", "uint8"),
]

# The corresponding nested array description (brief version):
testADescr2 = [
    ("x", "(2,)i4"),
    (
        "Info",
        [
            ("value", "()c16"),
            ("y2", "()f8"),
            (
                "Info2",
                [
                    ("name", "()S2"),
                    ("value", "(2,)c16"),
                    ("y3", "(2,)f8"),
                    ("z3", "(2,)i4"),
                ],
            ),
            ("name", "()S2"),
            ("z2", "()u1"),
        ],
    ),
    ("color", "()S2"),
    ("info", [("Name", "()S2"), ("Value", "()c16")]),
    ("y", "(2, 2)f8"),
    ("z", "()u1"),
]

# A nested array for testing:
testABuffer = [
    # x     Info    color info      y       z
    #       value y2 Info2      name z2         Name Value
    #                name   value    y3       z3
    (
        (3, 2),
        (6j, 6.0, ("nn", (6j, 4j), (6.0, 4.0), (1, 2)), "NN", 8),
        "cc",
        ("NN", 6j),
        ((6.0, 4.0), (6.0, 4.0)),
        8,
    ),
    (
        (4, 3),
        (7j, 7.0, ("oo", (7j, 5j), (7.0, 5.0), (2, 1)), "OO", 9),
        "dd",
        ("OO", 7j),
        ((7.0, 5.0), (7.0, 5.0)),
        9,
    ),
]
testAData = np.array(testABuffer, dtype=testADescr)
# The name of the column to be searched:
testCondCol = "Info/z2"
# The name of a nested column (it can not be searched):
testNestedCol = "Info"
# The condition to be applied on the column (all but the last row match it):
testCondition = "(2 < col) & (col < 9)"


def areDescriptionsEqual(desc1, desc2):
    """Are both `desc1` and `desc2` equivalent descriptions?

    The arguments may be description objects (``IsDescription``,
    ``Description``) or dictionaries.

    """

    if isinstance(desc1, tb.Col):
        # This is a rough comparison but it suffices here.
        return (
            desc1.type == desc2.type
            and desc2.dtype == desc2.dtype
            and desc1._v_pos == desc2._v_pos
            # and desc1.dflt == desc2.dflt)
            and common.areArraysEqual(desc1.dflt, desc2.dflt)
        )

    if hasattr(desc1, "_v_colobjects"):  # quacks like a Description
        cols1 = desc1._v_colobjects
    elif hasattr(desc1, "columns"):  # quacks like an IsDescription
        cols1 = desc1.columns
    else:  # hope it quacks like a dictionary
        cols1 = desc1

    if hasattr(desc2, "_v_colobjects"):  # quacks like a Description
        cols2 = desc2._v_colobjects
    elif hasattr(desc2, "columns"):  # quacks like an IsDescription
        cols2 = desc2.columns
    else:  # hope it quacks like a dictionary
        cols2 = desc2

    if len(cols1) != len(cols2):
        return False

    for colName, colobj1 in cols1.items():
        colobj2 = cols2[colName]
        if colName == "_v_pos":
            # The comparison may not be quite exhaustive!
            return colobj1 == colobj2
        if not areDescriptionsEqual(colobj1, colobj2):
            return False

    return True


# Test creating nested column descriptions
class DescriptionTestCase(common.PyTablesTestCase):
    _TestTDescr = TestTDescr
    _testADescr = testADescr
    _testADescr2 = testADescr2
    _testAData = testAData

    def test00_instance(self):
        """Creating an instance of a nested description."""

        self.assertTrue(
            areDescriptionsEqual(self._TestTDescr, self._TestTDescr()),
            "Table description does not match the given one.",
        )

    def test01_instance(self):
        """Checking attrs of an instance of a nested description."""

        descr = tb.description.Description(self._TestTDescr().columns)
        if common.verbose:
            print("Generated description:", descr._v_nested_descr)
            print("Should look like:", self._testADescr2)
        self.assertEqual(
            self._testADescr2,
            descr._v_nested_descr,
            "Description._v_nested_descr does not match.",
        )


# Test creating a nested table and opening it
class CreateTestCase(common.TempFileMixin, common.PyTablesTestCase):
    _TestTDescr = TestTDescr
    _testABuffer = testABuffer
    _testAData = testAData

    def _checkColumns(self, cols, desc):
        """Check that `cols` has all the accessors for `self._TestTDescr`."""

        # ``_desc`` is a leaf column and ``cols`` a ``Column``.
        if isinstance(desc, tb.Col):
            return isinstance(cols, tb.Column)

        # ``_desc`` is a description object and ``cols`` a ``Cols``.
        descColumns = desc._v_colobjects
        for colName in descColumns:
            if colName not in cols._v_colnames:
                return False
            if not self._checkColumns(
                cols._f_col(colName), descColumns[colName]
            ):
                return False

        return True

    def _checkDescription(self, table):
        """Check that description of `table` matches `self._TestTDescr`."""

        # Compare descriptions.
        self.assertTrue(
            areDescriptionsEqual(self._TestTDescr, table.description),
            "Table description does not match the given one.",
        )
        # Check access to columns.
        self._checkColumns(table.cols, table.description)

    def _checkColinstances(self, table):
        """Check that ``colinstances`` and ``cols`` of `table` match."""
        for colpathname in table.description._v_pathnames:
            self.assertTrue(
                table.colinstances[colpathname]
                is table.cols._f_col(colpathname)
            )

    def test00_create(self):
        """Creating a nested table."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        self._checkDescription(tbl)
        self._checkColinstances(tbl)

    def test01_open(self):
        """Opening a nested table."""

        self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        self._reopen()
        tbl = self.h5file.root.test
        self._checkDescription(tbl)
        self._checkColinstances(tbl)

    def test02_NestedRecArrayCompat(self):
        """Creating a compatible nested record array``."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        self.assertTrue(
            common.areArraysEqual(nrarr, self._testAData),
            "Can not create a compatible structured array.",
        )

    def test03_NRA(self):
        """Creating a table from a nested record array object."""

        tbl = self.h5file.create_table(
            "/", "test", self._testAData, title=self._getMethodName()
        )
        tbl.flush()
        readAData = tbl.read()
        if common.verbose:
            print("Read data:", readAData)
            print("Should look like:", self._testAData)
        self.assertTrue(
            common.areArraysEqual(self._testAData, readAData),
            "Written and read values differ.",
        )

    def test04_NRA2(self):
        """Creating a table from a generated nested record array object."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        readAData = tbl.read()

        tbl2 = self.h5file.create_table(
            "/", "test2", readAData, title=self._getMethodName()
        )
        readAData2 = tbl2.read()

        self.assertTrue(
            common.areArraysEqual(self._testAData, readAData2),
            "Written and read values differ.",
        )


# Test writing data in a nested table
class WriteTestCase(common.TempFileMixin, common.PyTablesTestCase):
    _TestTDescr = TestTDescr
    _testAData = testAData
    _testCondition = testCondition
    _testCondCol = testCondCol
    _testNestedCol = testNestedCol

    def _testCondVars(self, table):
        """Get condition variables for the given `table`."""
        return {"col": table.cols._f_col(self._testCondCol)}

    def _testNestedCondVars(self, table):
        """Get condition variables for the given `table`."""
        return {"col": table.cols._f_col(self._testNestedCol)}

    def _appendRow(self, row, index):
        """
        Append the `index`-th row in `self._testAData` to `row`.

        Values are set field-by-field (be it nested or not).
        """

        record = self._testAData[index]
        for fieldName in self._testAData.dtype.names:
            row[fieldName] = record[fieldName]
        row.append()

    def test00_append(self):
        """Appending a set of rows."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        readAData = tbl.read()
        self.assertTrue(
            common.areArraysEqual(self._testAData, readAData),
            "Written and read values differ.",
        )

    def test01_row(self):
        """Appending individual rows."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )

        row = tbl.row
        # Add the first row
        self._appendRow(row, 0)
        # Add the rest of the rows field by field.
        for i in range(1, len(self._testAData)):
            self._appendRow(row, i)
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        readAData = tbl.read()
        self.assertTrue(
            common.areArraysEqual(self._testAData, readAData),
            "Written and read values differ.",
        )

    def test02_where(self):
        """Searching nested data."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        searchedCoords = tbl.get_where_list(
            self._testCondition, self._testCondVars(tbl)
        )

        # All but the last row match the condition.
        searchedCoords.sort()
        self.assertEqual(
            searchedCoords.tolist(),
            list(range(len(self._testAData) - 1)),
            "Search returned incorrect results.",
        )

    def test02b_whereAppend(self):
        """Searching nested data and appending it to another table."""

        tbl1 = self.h5file.create_table(
            "/", "test1", self._TestTDescr, title=self._getMethodName()
        )
        tbl1.append(self._testAData)
        tbl1.flush()

        tbl2 = self.h5file.create_table(
            "/", "test2", self._TestTDescr, title=self._getMethodName()
        )
        tbl1.append_where(tbl2, self._testCondition, self._testCondVars(tbl1))

        if self.reopen:
            self._reopen()
            tbl1 = self.h5file.root.test1
            tbl2 = self.h5file.root.test2

        searchedCoords = tbl2.get_where_list(
            self._testCondition, self._testCondVars(tbl2)
        )

        # All but the last row match the condition.
        searchedCoords.sort()
        self.assertEqual(
            searchedCoords.tolist(),
            list(range(len(self._testAData) - 1)),
            "Search returned incorrect results.",
        )

    def test03_colscond(self):
        """Searching on a column with nested columns."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        self.assertRaises(
            TypeError,
            tbl.get_where_list,
            self._testCondition,
            self._testNestedCondVars(tbl),
        )

    def test04_modifyColumn(self):
        """Modifying one single nested column (modify_column)."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        tbl.flush()

        nColumn = self._testNestedCol
        # Get the nested column data and swap the first and last rows.
        raTable = self._testAData.copy()
        raColumn = raTable[nColumn]
        # The next will not work until NestedRecords supports copies
        raColumn[0], raColumn[-1] = (raColumn[-1], raColumn[0])

        # Write the resulting column and re-read the whole table.
        tbl.modify_column(colname=nColumn, column=raColumn)
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        raReadTable = tbl.read()
        if common.verbose:
            print("Table read:", raReadTable)
            print("Should look like:", raTable)

        # Compare it to the written one.
        self.assertTrue(
            common.areArraysEqual(raTable, raReadTable),
            "Written and read values differ.",
        )

    def test05a_modifyColumns(self):
        """Modifying one nested column (modify_columns)."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        tbl.flush()

        nColumn = self._testNestedCol
        # Get the nested column data and swap the first and last rows.
        raTable = self._testAData.copy()
        raColumn = raTable[nColumn]
        raColumn[0], raColumn[-1] = (raColumn[-1].copy(), raColumn[0].copy())
        newdtype = np.dtype([(nColumn, raTable.dtype.fields[nColumn][0])])
        self.assertIsNotNone(newdtype)

        # Write the resulting column and re-read the whole table.
        tbl.modify_columns(names=[nColumn], columns=raColumn)
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        raReadTable = tbl.read()
        if common.verbose:
            print("Table read:", raReadTable)
            print("Should look like:", raTable)

        # Compare it to the written one.
        self.assertTrue(
            common.areArraysEqual(raTable, raReadTable),
            "Written and read values differ.",
        )

    def test05b_modifyColumns(self):
        """Modifying two nested columns (modify_columns)."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        tbl.flush()

        # Get the nested column data and swap the first and last rows.
        colnames = ["x", "color"]  # Get the first two columns
        raCols = np.rec.fromarrays(
            [self._testAData["x"].copy(), self._testAData["color"].copy()],
            dtype=[("x", "(2,)i4"), ("color", "S2")],
        )
        # descr=tbl.description._v_nested_descr[0:2])
        # or...
        # names=tbl.description._v_nested_names[0:2],
        # formats=tbl.description._v_nested_formats[0:2])
        raCols[0], raCols[-1] = (raCols[-1].copy(), raCols[0].copy())

        # Write the resulting columns
        tbl.modify_columns(names=colnames, columns=raCols)
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        # Re-read the appropriate columns
        raCols2 = np.rec.fromarrays(
            [tbl.cols._f_col("x"), tbl.cols._f_col("color")],
            dtype=raCols.dtype,
        )
        if common.verbose:
            print("Table read:", raCols2)
            print("Should look like:", raCols)

        # Compare it to the written one.
        self.assertTrue(
            common.areArraysEqual(raCols, raCols2),
            "Written and read values differ.",
        )

    def test06_modifyRows(self):
        """Checking modifying several rows at once (using nested rec array)"""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        tbl.flush()

        # Get the nested record and swap the first and last rows.
        raTable = self._testAData.copy()
        raTable[0], raTable[-1] = (raTable[-1].copy(), raTable[0].copy())

        # Write the resulting nested record and re-read the whole table.
        tbl.modify_rows(start=0, stop=2, rows=raTable)
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        raReadTable = tbl.read()
        if common.verbose:
            print("Table read:", raReadTable)
            print("Should look like:", raTable)

        # Compare it to the written one.
        self.assertTrue(
            common.areArraysEqual(raTable, raReadTable),
            "Written and read values differ.",
        )

    def test07_index(self):
        """Checking indexes of nested columns."""

        tbl = self.h5file.create_table(
            "/",
            "test",
            self._TestTDescr,
            title=self._getMethodName(),
            expectedrows=minRowIndex * 2,
        )
        for i in range(minRowIndex):
            tbl.append(self._testAData)
        tbl.flush()
        coltoindex = tbl.cols._f_col(self._testCondCol)
        indexrows = coltoindex.create_index()
        self.assertIsNotNone(indexrows)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test
            coltoindex = tbl.cols._f_col(self._testCondCol)

        if common.verbose:
            print("Number of written rows:", tbl.nrows)
            print("Number of indexed rows:", coltoindex.index.nelements)

        # Check indexing flags:
        self.assertEqual(tbl.indexed, True, "Table not indexed")
        self.assertNotEqual(coltoindex.index, None, "Column not indexed")
        self.assertTrue(
            tbl.colindexed[self._testCondCol], "Column not indexed"
        )
        # Do a look-up for values
        searchedCoords = tbl.get_where_list(
            self._testCondition, self._testCondVars(tbl)
        )
        searchedCoords.sort()

        expectedCoords = np.arange(0, minRowIndex * 2, 2, tb.utils.SizeType)
        if common.verbose:
            print("Searched coords:", searchedCoords)
            print("Expected coords:", expectedCoords)
        # All even rows match the condition.
        self.assertEqual(
            searchedCoords.tolist(),
            expectedCoords.tolist(),
            "Search returned incorrect results.",
        )

    def test08_setNestedField(self):
        """Checking modifying a nested field via natural naming."""
        # See ticket #93 (http://www.pytables.org/trac/ticket/93).

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)
        tbl.flush()

        oldvalue = tbl.cols.Info.z2[0]
        tbl.cols.Info.z2[0] = oldvalue + 1
        tbl.flush()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        newvalue = tbl.cols.Info.z2[0]
        self.assertEqual(newvalue, oldvalue + 1)


class WriteNoReopen(WriteTestCase):
    reopen = 0


class WriteReopen(WriteTestCase):
    reopen = 1


class ReadTestCase(common.TempFileMixin, common.PyTablesTestCase):
    _TestTDescr = TestTDescr
    _testABuffer = testABuffer
    _testAData = testAData
    _testNestedCol = testNestedCol

    def test00a_repr(self):
        """Checking representation of a nested Table."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title="test00"
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        if common.verbose:
            print("str(tbl)-->", str(tbl))
            print("repr(tbl)-->", repr(tbl))

        self.assertEqual(
            str(tbl), f"/test (Table({np.int64(2)!r},)) {np.str_('test00')!r}"
        )
        tblrepr = repr(tbl)
        # Remove the platform-dependent information (i.e. byteorder)
        tblrepr = "\n".join(tblrepr.split("\n")[:-2]) + "\n"
        template = f"""/test (Table({np.int64(2)!r},)) {np.str_('test00')!r}
  description := {{
  "x": Int32Col(shape=({np.int64(2)!r},), dflt={np.int32(0)!r}, pos=0),
  "Info": {{
    "value": ComplexCol(itemsize=16, shape=(), dflt={np.complex128(0j)!r}, pos=0),
    "y2": Float64Col(shape=(), dflt={np.float64(1.0)!r}, pos=1),
    "Info2": {{
      "name": StringCol(itemsize=2, shape=(), dflt={np.bytes_(b'')!r}, pos=0),
      "value": ComplexCol(itemsize=16, shape=({np.int64(2)!r},), dflt={np.complex128(0j)!r}, pos=1),
      "y3": Time64Col(shape=({np.int64(2)!r},), dflt={np.float64(1.0)!r}, pos=2),
      "z3": EnumCol(enum=Enum({{%(value)s}}), dflt='%(default)s', base=Int32Atom(shape=(), dflt={np.int32(0)!r}), shape=({np.int64(2)!r},), pos=3)}},
    "name": StringCol(itemsize=2, shape=(), dflt={np.bytes_(b'')!r}, pos=3),
    "z2": UInt8Col(shape=(), dflt={np.uint8(1)!r}, pos=4)}},
  "color": StringCol(itemsize=2, shape=(), dflt={np.bytes_(b' ')!r}, pos=2),
  "info": {{
    "Name": StringCol(itemsize=2, shape=(), dflt={np.bytes_(b'')!r}, pos=0),
    "Value": ComplexCol(itemsize=16, shape=(), dflt={np.complex128(0j)!r}, pos=1)}},
  "y": Float64Col(shape=({np.int64(2)!r}, {np.int64(2)!r}), dflt={np.float64(1.0)!r}, pos=4),
  "z": UInt8Col(shape=(), dflt={np.uint8(1)!r}, pos=5)}}
"""

        # The problem here is that the order in which items are stored in a
        # dict can't be assumed to be stable.
        # From python 3.3 on it is actually no more stable since the
        # "Hash randomization" feature is enable by default.
        #
        # For this reason we generate a representation string for each of the
        # permutations of the Enum items.
        #
        # Also the default value of enum types is not preserved in HDF5.
        # It is assumed that the default value is the first one in the array
        # of Enum names and hence it is also affected by the issue related to
        # the "Hash randomization" feature.
        #
        # Also in this case it is generated a representation string for each
        # of the possible default values.
        enums = [
            ", ".join(items)
            for items in itertools.permutations(
                (
                    f"'r': {np.int32(4)!r}",
                    f"'b': {np.int32(1)!r}",
                    f"'g': {np.int32(2)!r}",
                )
                if self.reopen
                else ("'r': 4", "'b': 1", "'g': 2")
            )
        ]
        defaults = ("r", "b", "g")
        values = [
            template % {"value": v, "default": d}
            for v, d in itertools.product(enums, defaults)
        ]
        self.assertIn(tblrepr, values)

    def test00b_repr(self):
        """Checking representation of a root Column."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title="test00"
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        if common.verbose:
            print("str(tbl.cols.y)-->'%s'" % str(tbl.cols.y))
            print("repr(tbl.cols.y)-->'%s'" % repr(tbl.cols.y))

        self.assertEqual(
            str(tbl.cols.y),
            f"/test.cols.y (Column({np.int64(2)!r}, 2, 2), float64, idx=None)",
        )
        self.assertEqual(
            repr(tbl.cols.y),
            f"/test.cols.y (Column({np.int64(2)!r}, 2, 2), float64, idx=None)",
        )

    def test00c_repr(self):
        """Checking representation of a nested Column."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title="test00"
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        if common.verbose:
            print("str(tbl.cols.Info.z2)-->'%s'" % str(tbl.cols.Info.z2))
            print("repr(tbl.cols.Info.z2)-->'%s'" % repr(tbl.cols.Info.z2))

        self.assertEqual(
            str(tbl.cols.Info.z2),
            f"/test.cols.Info.z2 (Column({np.int64(2)!r},), uint8, idx=None)",
        )
        self.assertEqual(
            repr(tbl.cols.Info.z2),
            f"/test.cols.Info.z2 (Column({np.int64(2)!r},), uint8, idx=None)",
        )

    def test01_read(self):
        """Checking Table.read with subgroups with a range index with step."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.rec.array(
            testABuffer, dtype=tbl.description._v_nested_descr
        )
        tblcols = tbl.read(start=0, step=2, field="Info")
        nrarrcols = nrarr["Info"][0::2]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test01_read_out_arg(self):
        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.rec.array(
            testABuffer, dtype=tbl.description._v_nested_descr
        )
        # When reading an entire nested column, the output array must contain
        # all fields in the table.  The output buffer will contain the contents
        # of all fields.  The selected column alone will be returned from the
        # method call.
        all_cols = np.empty(1, tbl.dtype)
        tblcols = tbl.read(start=0, step=2, field="Info", out=all_cols)
        nrarrcols = nrarr["Info"][0::2]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )
        self.assertTrue(
            common.areArraysEqual(nrarr[0::2], all_cols),
            "Output buffer does not match full table.",
        )

    def test02_read(self):
        """Checking Table.read with a nested Column."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        tblcols = tbl.read(start=0, step=2, field="Info/value")
        nrarr = np.rec.array(
            testABuffer, dtype=tbl.description._v_nested_descr
        )
        nrarrcols = nrarr["Info"]["value"][0::2]
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test02_read_out_arg(self):
        """Checking Table.read with a nested Column."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        tblcols = np.empty(1, dtype="c16")
        tbl.read(start=0, step=2, field="Info/value", out=tblcols)
        nrarr = np.rec.array(
            testABuffer, dtype=tbl.description._v_nested_descr
        )
        nrarrcols = nrarr["Info"]["value"][0::2]
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )


class ReadNoReopen(ReadTestCase):
    reopen = 0


class ReadReopen(ReadTestCase):
    reopen = 1


# Checking the Table.Cols accessor
class ColsTestCase(common.TempFileMixin, common.PyTablesTestCase):
    _TestTDescr = TestTDescr
    _testABuffer = testABuffer
    _testAData = testAData
    _testNestedCol = testNestedCol

    def test00a_repr(self):
        """Checking string representation of Cols."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title="test00"
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        if common.verbose:
            print("str(tbl.cols)-->", str(tbl.cols))
            print("repr(tbl.cols)-->", repr(tbl.cols))

        self.assertEqual(str(tbl.cols), "/test.cols (Cols), 6 columns")
        try:
            self.assertEqual(
                repr(tbl.cols),
                """/test.cols (Cols), 6 columns
  x (Column(0, 2), ('int32',(2,)))
  Info (Cols(), Description)
  color (Column(0,), |S2)
  info (Cols(), Description)
  y (Column(0, 2, 2), ('float64',(2, 2)))
  z (Column(0,), uint8)
""",
            )
        except AssertionError:
            self.assertEqual(
                repr(tbl.cols),
                f"""/test.cols (Cols), 6 columns
  x (Column({np.int64(0)!r}, 2), ('{np.int32(0).dtype.str}', (2,)))
  Info (Cols(), Description)
  color (Column({np.int64(0)!r},), |S2)
  info (Cols(), Description)
  y (Column({np.int64(0)!r}, 2, 2), ('{np.float64(0).dtype.str}', (2, 2)))
  z (Column({np.int64(0)!r},), uint8)
""",
            )

    def test00b_repr(self):
        """Checking string representation of nested Cols."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        if common.verbose:
            print("str(tbl.cols.Info)-->", str(tbl.cols.Info))
            print("repr(tbl.cols.Info)-->", repr(tbl.cols.Info))

        self.assertEqual(
            str(tbl.cols.Info), "/test.cols.Info (Cols), 5 columns"
        )
        self.assertEqual(
            repr(tbl.cols.Info),
            f"""/test.cols.Info (Cols), 5 columns
  value (Column({np.int64(0)!r},), complex128)
  y2 (Column({np.int64(0)!r},), float64)
  Info2 (Cols(), Description)
  name (Column({np.int64(0)!r},), |S2)
  z2 (Column({np.int64(0)!r},), uint8)
""",
        )

    def test01a_f_col(self):
        """Checking cols._f_col() with a subgroup."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        tblcol = tbl.cols._f_col(self._testNestedCol)
        if common.verbose:
            print("Column group name:", tblcol._v_desc._v_pathname)
        self.assertEqual(
            tblcol._v_desc._v_pathname,
            self._testNestedCol,
            "Column group name doesn't match.",
        )

    def test01b_f_col(self):
        """Checking cols._f_col() with a column."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        tblcol = tbl.cols._f_col(self._testNestedCol + "/name")
        if common.verbose:
            print("Column name:", tblcol.name)
        self.assertEqual(tblcol.name, "name", "Column name doesn't match.")

    def test01c_f_col(self):
        """Checking cols._f_col() with a nested subgroup."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )

        tblcol = tbl.cols._f_col(self._testNestedCol + "/Info2")
        if common.verbose:
            print("Column group name:", tblcol._v_desc._v_pathname)
        self.assertEqual(
            tblcol._v_desc._v_pathname,
            self._testNestedCol + "/Info2",
            "Column group name doesn't match.",
        )

    def test02a__len__(self):
        """Checking cols.__len__() in root level."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        length = len(tbl.cols)
        if common.verbose:
            print("Column group length:", length)
        self.assertEqual(
            length, len(tbl.colnames), "Column group length doesn't match."
        )

    def test02b__len__(self):
        """Checking cols.__len__() in subgroup level."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        length = len(tbl.cols.Info)
        if common.verbose:
            print("Column group length:", length)
        self.assertEqual(
            length,
            len(tbl.cols.Info._v_colnames),
            "Column group length doesn't match.",
        )

    def test03a__getitem__(self):
        """Checking cols.__getitem__() with a single index."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols[1]
        nrarrcols = nrarr[1]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test03b__getitem__(self):
        """Checking cols.__getitem__() with a range index."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols[0:2]
        nrarrcols = nrarr[0:2]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test03c__getitem__(self):
        """Checking cols.__getitem__() with a range index with step."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols[0::2]
        nrarrcols = nrarr[0::2]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test04a__getitem__(self):
        """Checking cols.__getitem__() with subgroups with a single index."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols._f_col("Info")[1]
        nrarrcols = nrarr["Info"][1]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test04b__getitem__(self):
        """Checking cols.__getitem__() with subgroups with a range index."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols._f_col("Info")[0:2]
        nrarrcols = nrarr["Info"][0:2]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test04c__getitem__(self):
        """Checking cols.__getitem__() with subgroups with a range index with
        step."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols._f_col("Info")[0::2]
        nrarrcols = nrarr["Info"][0::2]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test05a__getitem__(self):
        """Checking cols.__getitem__() with a column with a single index."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols._f_col("Info/value")[1]
        nrarrcols = nrarr["Info"]["value"][1]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertEqual(
            nrarrcols, tblcols, "Original array are retrieved doesn't match."
        )

    def test05b__getitem__(self):
        """Checking cols.__getitem__() with a column with a range index."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols._f_col("Info/value")[0:2]
        nrarrcols = nrarr["Info"]["value"][0:2]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test05c__getitem__(self):
        """Checking cols.__getitem__() with a column with a range index with
        step."""

        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        tblcols = tbl.cols._f_col("Info/value")[0::2]
        nrarrcols = nrarr["Info"]["value"][0::2]
        if common.verbose:
            print("Read cols:", tblcols)
            print("Should look like:", nrarrcols)
        self.assertTrue(
            common.areArraysEqual(nrarrcols, tblcols),
            "Original array are retrieved doesn't match.",
        )

    def test_01a__iter__(self):
        tbl = self.h5file.create_table(
            "/", "test", self._TestTDescr, title=self._getMethodName()
        )
        tbl.append(self._testAData)

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        nrarr = np.array(testABuffer, dtype=tbl.description._v_nested_descr)
        row_num = 0
        for item in tbl.cols.Info.value:
            self.assertEqual(item, nrarr["Info"]["value"][row_num])
            row_num += 1
        self.assertEqual(row_num, len(nrarr))


class ColsNoReopen(ColsTestCase):
    reopen = 0


class ColsReopen(ColsTestCase):
    reopen = 1


class Nested(tb.IsDescription):
    uid = tb.IntCol(pos=1)
    value = tb.FloatCol(pos=2)


class A_Candidate(tb.IsDescription):
    nested1 = Nested()
    nested2 = Nested()


class B_Candidate(tb.IsDescription):
    nested1 = Nested
    nested2 = Nested


class C_Candidate(tb.IsDescription):
    nested1 = Nested()
    nested2 = Nested


Dnested = {
    "uid": tb.IntCol(pos=1),
    "value": tb.FloatCol(pos=2),
}

D_Candidate = {
    "nested1": Dnested,
    "nested2": Dnested,
}

E_Candidate = {
    "nested1": Nested,
    "nested2": Dnested,
}

F_Candidate = {
    "nested1": Nested(),
    "nested2": Dnested,
}

# Checking several nested columns declared in the same way


class SameNestedTestCase(common.TempFileMixin, common.PyTablesTestCase):
    correct_names = [
        "",  # The root of columns
        "nested1",
        "nested1/uid",
        "nested1/value",
        "nested2",
        "nested2/uid",
        "nested2/value",
    ]

    def test01a(self):
        """Checking same nested columns (instance flavor)."""

        tbl = self.h5file.create_table(
            "/", "test", A_Candidate, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        names = [
            col._v_pathname for col in tbl.description._f_walk(type="All")
        ]
        if common.verbose:
            print("Pathnames of columns:", names)
            print("Should look like:", self.correct_names)
        self.assertEqual(
            names, self.correct_names, "Column nested names doesn't match."
        )

    def test01b(self):
        """Checking same nested columns (class flavor)."""

        tbl = self.h5file.create_table(
            "/", "test", B_Candidate, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        names = [
            col._v_pathname for col in tbl.description._f_walk(type="All")
        ]
        if common.verbose:
            print("Pathnames of columns:", names)
            print("Should look like:", self.correct_names)
        self.assertEqual(
            names, self.correct_names, "Column nested names doesn't match."
        )

    def test01c(self):
        """Checking same nested columns (mixed instance/class flavor)."""

        tbl = self.h5file.create_table(
            "/", "test", C_Candidate, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        names = [
            col._v_pathname for col in tbl.description._f_walk(type="All")
        ]
        if common.verbose:
            print("Pathnames of columns:", names)
            print("Should look like:", self.correct_names)
        self.assertEqual(
            names, self.correct_names, "Column nested names doesn't match."
        )

    def test01d(self):
        """Checking same nested columns (dictionary flavor)."""

        tbl = self.h5file.create_table(
            "/", "test", D_Candidate, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        names = [
            col._v_pathname for col in tbl.description._f_walk(type="All")
        ]
        if common.verbose:
            print("Pathnames of columns:", names)
            print("Should look like:", self.correct_names)
        self.assertEqual(
            names, self.correct_names, "Column nested names doesn't match."
        )

    def test01e(self):
        """Checking same nested columns (mixed dictionary/class flavor)."""

        tbl = self.h5file.create_table(
            "/", "test", E_Candidate, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        names = [
            col._v_pathname for col in tbl.description._f_walk(type="All")
        ]
        if common.verbose:
            print("Pathnames of columns:", names)
            print("Should look like:", self.correct_names)
        self.assertEqual(
            names, self.correct_names, "Column nested names doesn't match."
        )

    def test01f(self):
        """Checking same nested columns (mixed dictionary/instance flavor)."""

        tbl = self.h5file.create_table(
            "/", "test", F_Candidate, title=self._getMethodName()
        )

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test

        names = [
            col._v_pathname for col in tbl.description._f_walk(type="All")
        ]
        if common.verbose:
            print("Pathnames of columns:", names)
            print("Should look like:", self.correct_names)
        self.assertEqual(
            names, self.correct_names, "Column nested names doesn't match."
        )

    def test02a(self):
        """Indexing two simple columns under the same nested column."""

        desc = {"nested": {"i1": tb.Int32Col(), "i2": tb.Int32Col()}}

        i1 = "nested/i1"
        i2 = "nested/i2"
        tbl = self.h5file.create_table(
            "/", "test", desc, title=self._getMethodName()
        )

        row = tbl.row
        for i in range(1000):
            row[i1] = i
            row[i2] = i * 2
            row.append()
        tbl.flush()

        cols = {
            "i1": tbl.cols.nested.i1,
            "i2": tbl.cols.nested.i2,
        }
        cols["i1"].create_index()
        cols["i2"].create_index()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test
            # Redefine the cols dictionary
            cols = {
                "i1": tbl.cols.nested.i1,
                "i2": tbl.cols.nested.i2,
            }

        i1res = [r[i1] for r in tbl.where("i1 < 10", cols)]
        i2res = [r[i2] for r in tbl.where("i2 < 10", cols)]

        if common.verbose:
            print("Retrieved values (i1):", i1res)
            print("Should look like:", list(range(10)))
            print("Retrieved values (i2):", i2res)
            print("Should look like:", list(range(0, 10, 2)))

        self.assertEqual(
            i1res,
            list(range(10)),
            "Select for nested column (i1) doesn't match.",
        )
        self.assertEqual(
            i2res,
            list(range(0, 10, 2)),
            "Select for nested column (i2) doesn't match.",
        )

    def test02b(self):
        """Indexing two simple columns under the same (very) nested column."""

        desc = {
            "nested1": {
                "nested2": {
                    "nested3": {"i1": tb.Int32Col(), "i2": tb.Int32Col()}
                }
            }
        }

        i1 = "nested1/nested2/nested3/i1"
        i2 = "nested1/nested2/nested3/i2"

        tbl = self.h5file.create_table(
            "/", "test", desc, title=self._getMethodName()
        )

        row = tbl.row
        for i in range(1000):
            row[i1] = i
            row[i2] = i * 2
            row.append()
        tbl.flush()

        cols = {
            "i1": tbl.cols.nested1.nested2.nested3.i1,
            "i2": tbl.cols.nested1.nested2.nested3.i2,
        }
        cols["i1"].create_index()
        cols["i2"].create_index()

        if self.reopen:
            self._reopen()
            tbl = self.h5file.root.test
            # Redefine the cols dictionary
            cols = {
                "i1": tbl.cols.nested1.nested2.nested3.i1,
                "i2": tbl.cols.nested1.nested2.nested3.i2,
            }

        i1res = [r[i1] for r in tbl.where("i1 < 10", cols)]
        i2res = [r[i2] for r in tbl.where("i2 < 10", cols)]

        if common.verbose:
            print("Retrieved values (i1):", i1res)
            print("Should look like:", list(range(10)))
            print("Retrieved values (i2):", i2res)
            print("Should look like:", list(range(0, 10, 2)))

        self.assertEqual(
            i1res,
            list(range(10)),
            "Select for nested column (i1) doesn't match.",
        )
        self.assertEqual(
            i2res,
            list(range(0, 10, 2)),
            "Select for nested column (i2) doesn't match.",
        )


class SameNestedNoReopen(SameNestedTestCase):
    reopen = 0


class SameNestedReopen(SameNestedTestCase):
    reopen = 1


class NestedTypesWithGaps(common.TestFileMixin, common.PyTablesTestCase):
    h5fname = common.test_filename("nested-type-with-gaps.h5")

    correct_descr = f"""{{
  "float": Float32Col(shape=(), dflt={np.float32(0.0)!r}, pos=0),
  "compound": {{
    "char": Int8Col(shape=(), dflt={np.int8(0)!r}, pos=0),
    "double": Float64Col(shape=(), dflt={np.float64(0.0)!r}, pos=1)}}}}"""

    def test01(self):
        """Opening a table with nested types with gaps."""

        tbl = self.h5file.get_node("/nestedtype")
        type_descr = repr(tbl.description)
        if common.verbose:
            print("Type size with gaps:", tbl.description._v_itemsize)
            print("And should be: 16")
            print("Representation of the nested type:\n", type_descr)
            print("And should be:\n", self.correct_descr)
            print("Here are the offsets: ", tbl.description._v_offsets)

        self.assertEqual(tbl.description._v_itemsize, 16)
        self.assertEqual(type_descr, self.correct_descr)

        if common.verbose:
            print("Great!  Nested types with gaps recognized correctly.")


def suite():
    """Return a test suite consisting of all the test cases in the module."""

    theSuite = common.unittest.TestSuite()
    niter = 1
    # common.heavy = 1  # uncomment this only for testing purposes

    for i in range(niter):
        theSuite.addTest(common.make_suite(DescriptionTestCase))
        theSuite.addTest(common.make_suite(CreateTestCase))
        theSuite.addTest(common.make_suite(WriteNoReopen))
        theSuite.addTest(common.make_suite(WriteReopen))
        theSuite.addTest(common.make_suite(ColsNoReopen))
        theSuite.addTest(common.make_suite(ColsReopen))
        theSuite.addTest(common.make_suite(ReadNoReopen))
        theSuite.addTest(common.make_suite(ReadReopen))
        theSuite.addTest(common.make_suite(SameNestedNoReopen))
        theSuite.addTest(common.make_suite(SameNestedReopen))
        theSuite.addTest(common.make_suite(NestedTypesWithGaps))

    return theSuite


if __name__ == "__main__":
    common.parse_argv(sys.argv)
    common.print_versions()
    common.unittest.main(defaultTest="suite")



"""Utility to helps the migration from PyTables 2.x APIs to 3.x APIs.

The new API is PEP 8 compliant.

"""

import re
import sys
import argparse
from pathlib import Path

old2newnames = dict(  # noqa: C406
    [
        # from __init__.py
        ("hdf5Version", "hdf5_version"),  # data
        # from array.py
        ("parentNode", "parentnode"),  # kwarg
        ("getEnum", "get_enum"),
        ("_initLoop", "_init_loop"),
        ("_fancySelection", "_fancy_selection"),
        ("_checkShape", "_check_shape"),
        ("_readSlice", "_read_slice"),
        ("_readCoords", "_read_coords"),
        ("_readSelection", "_read_selection"),
        ("_writeSlice", "_write_slice"),
        ("_writeCoords", "_write_coords"),
        ("_writeSelection", "_write_selection"),
        ("_g_copyWithStats", "_g_copy_with_stats"),
        ("_c_classId", "_c_classid"),  # attr
        # from atom.py
        ("_checkBase", "_checkbase"),
        # from attributeset.py
        ("newSet", "newset"),  # kwarg
        ("copyClass", "copyclass"),  # kwarg
        ("_g_updateNodeLocation", "_g_update_node_location"),
        ("_g_logAdd", "_g_log_add"),
        ("_g_delAndLog", "_g_del_and_log"),
        ("_v__nodeFile", "_v__nodefile"),  # attr (private)
        ("_v__nodePath", "_v__nodepath"),  # attr (private)
        # from carray.py
        # ('parentNode', 'parentnode'),                       # kwarg
        # from description.py
        ("_g_setNestedNamesDescr", "_g_set_nested_names_descr"),
        ("_g_setPathNames", "_g_set_path_names"),
        ("_v_colObjects", "_v_colobjects"),  # attr
        ("_v_nestedFormats", "_v_nested_formats"),  # attr
        ("_v_nestedNames", "_v_nested_names"),  # attr
        ("_v_nestedDescr", "_v_nested_descr"),  # attr
        ("getColsInOrder", "get_cols_in_order"),
        ("joinPaths", "join_paths"),
        ("metaIsDescription", "MetaIsDescription"),
        # from earray.py
        # ('parentNode', 'parentnode'),                       # kwarg
        ("_checkShapeAppend", "_check_shape_append"),
        # from expression.py
        ("_exprvarsCache", "_exprvars_cache"),  # attr (private)
        ("_requiredExprVars", "_required_expr_vars"),
        ("setInputsRange", "set_inputs_range"),
        ("setOutput", "set_output"),
        ("setOutputRange", "set_output_range"),
        # from file.py
        ("_opToCode", "_op_to_code"),  # data (private)
        ("_codeToOp", "_code_to_op"),  # data (private)
        ("_transVersion", "_trans_version"),  # data (private)
        ("_transGroupParent", "_trans_group_parent"),  # data (private)
        ("_transGroupName", "_trans_group_name"),  # data (private)
        ("_transGroupPath", "_trans_group_path"),  # data (private)
        ("_actionLogParent", "_action_log_parent"),  # data (private)
        ("_actionLogName", "_action_log_name"),  # data (private)
        ("_actionLogPath", "_action_log_path"),  # data (private)
        ("_transParent", "_trans_parent"),  # data (private)
        ("_transName", "_trans_name"),  # data (private)
        ("_transPath", "_trans_path"),  # data (private)
        ("_shadowParent", "_shadow_parent"),  # data (private)
        ("_shadowName", "_shadow_name"),  # data (private)
        ("_shadowPath", "_shadow_path"),  # data (private)
        ("copyFile", "copy_file"),
        ("openFile", "open_file"),
        ("_getValueFromContainer", "_get_value_from_container"),
        ("__getRootGroup", "__get_root_group"),
        ("rootUEP", "root_uep"),  # attr
        ("_getOrCreatePath", "_get_or_create_path"),
        ("_createPath", "_create_path"),
        ("createGroup", "create_group"),
        ("createTable", "create_table"),
        ("createArray", "create_array"),
        ("createCArray", "create_carray"),
        ("createEArray", "create_earray"),
        ("createVLArray", "create_vlarray"),
        ("createHardLink", "create_hard_link"),
        ("createSoftLink", "create_soft_link"),
        ("createExternalLink", "create_external_link"),
        ("_getNode", "_get_node"),
        ("getNode", "get_node"),
        ("isVisibleNode", "is_visible_node"),
        ("renameNode", "rename_node"),
        ("moveNode", "move_node"),
        ("copyNode", "copy_node"),
        ("removeNode", "remove_node"),
        ("getNodeAttr", "get_node_attr"),
        ("setNodeAttr", "set_node_attr"),
        ("delNodeAttr", "del_node_attr"),
        ("copyNodeAttrs", "copy_node_attrs"),
        ("copyChildren", "copy_children"),
        ("listNodes", "list_nodes"),
        ("iterNodes", "iter_nodes"),
        ("walkNodes", "walk_nodes"),
        ("walkGroups", "walk_groups"),
        ("_checkOpen", "_check_open"),
        ("_isWritable", "_iswritable"),
        ("_checkWritable", "_check_writable"),
        ("_checkGroup", "_check_group"),
        ("isUndoEnabled", "is_undo_enabled"),
        ("_checkUndoEnabled", "_check_undo_enabled"),
        ("_createTransactionGroup", "_create_transaction_group"),
        ("_createTransaction", "_create_transaction"),
        ("_createMark", "_create_mark"),
        ("enableUndo", "enable_undo"),
        ("disableUndo", "disable_undo"),
        ("_getMarkID", "_get_mark_id"),
        ("_getFinalAction", "_get_final_action"),
        ("getCurrentMark", "get_current_mark"),
        ("_updateNodeLocations", "_update_node_locations"),
        # from group.py
        # ('parentNode', 'parentnode'),                       # kwarg
        # ('ptFile', 'ptfile'),                               # kwarg
        ("_getValueFromContainer", "_get_value_from_container"),
        ("_g_postInitHook", "_g_post_init_hook"),
        ("_g_getChildGroupClass", "_g_get_child_group_class"),
        ("_g_getChildLeafClass", "_g_get_child_leaf_class"),
        ("_g_addChildrenNames", "_g_add_children_names"),
        ("_g_checkHasChild", "_g_check_has_child"),
        ("_f_walkNodes", "_f_walknodes"),
        ("_g_widthWarning", "_g_width_warning"),
        ("_g_refNode", "_g_refnode"),
        ("_g_unrefNode", "_g_unrefnode"),
        ("_g_copyChildren", "_g_copy_children"),
        ("_f_getChild", "_f_get_child"),
        ("_f_listNodes", "_f_list_nodes"),
        ("_f_iterNodes", "_f_iter_nodes"),
        ("_f_walkGroups", "_f_walk_groups"),
        ("_g_closeDescendents", "_g_close_descendents"),
        ("_f_copyChildren", "_f_copy_children"),
        ("_v_maxGroupWidth", "_v_max_group_width"),  # attr
        ("_v_objectID", "_v_objectid"),  # attr
        ("_g_loadChild", "_g_load_child"),
        ("childName", "childname"),  # ???
        ("_c_shadowNameRE", "_c_shadow_name_re"),  # attr (private)
        # from hdf5extension.p{yx,xd}
        ("hdf5Extension", "hdf5extension"),
        ("_getFileId", "_get_file_id"),
        ("_flushFile", "_flush_file"),
        ("_closeFile", "_close_file"),
        ("_g_listAttr", "_g_list_attr"),
        ("_g_setAttr", "_g_setattr"),
        ("_g_getAttr", "_g_getattr"),
        ("_g_listGroup", "_g_list_group"),
        ("_g_getGChildAttr", "_g_get_gchild_attr"),
        ("_g_getLChildAttr", "_g_get_lchild_attr"),
        ("_g_flushGroup", "_g_flush_group"),
        ("_g_closeGroup", "_g_close_group"),
        ("_g_moveNode", "_g_move_node"),
        ("_convertTime64", "_convert_time64"),
        ("_createArray", "_create_array"),
        ("_createCArray", "_create_carray"),
        ("_openArray", "_open_array"),
        ("_readArray", "_read_array"),
        ("_g_readSlice", "_g_read_slice"),
        ("_g_readCoords", "_g_read_coords"),
        ("_g_readSelection", "_g_read_selection"),
        ("_g_writeSlice", "_g_write_slice"),
        ("_g_writeCoords", "_g_write_coords"),
        ("_g_writeSelection", "_g_write_selection"),
        # from idxutils.py
        ("calcChunksize", "calc_chunksize"),
        ("infinityF", "infinityf"),  # data
        ("infinityMap", "infinitymap"),  # data
        ("infType", "inftype"),
        ("StringNextAfter", "string_next_after"),
        ("IntTypeNextAfter", "int_type_next_after"),
        ("BoolTypeNextAfter", "bool_type_next_after"),
        # from index.py
        # ('parentNode', 'parentnode'),                       # kwarg
        ("defaultAutoIndex", "default_auto_index"),  # data
        ("defaultIndexFilters", "default_index_filters"),  # data
        ("_tableColumnPathnameOfIndex", "_table_column_pathname_of_index"),
        ("_is_CSI", "_is_csi"),
        ("is_CSI", "is_csi"),  # property
        ("appendLastRow", "append_last_row"),
        ("read_sliceLR", "read_slice_lr"),
        ("readSorted", "read_sorted"),
        ("readIndices", "read_indices"),
        ("_processRange", "_process_range"),
        ("searchLastRow", "search_last_row"),
        ("getLookupRange", "get_lookup_range"),
        ("_g_checkName", "_g_check_name"),
        # from indexes.py
        # ('parentNode', 'parentnode'),                       # kwarg
        ("_searchBin", "_search_bin"),
        # from indexesextension
        ("indexesExtension", "indexesextension"),
        ("initRead", "initread"),
        ("readSlice", "read_slice"),
        ("_readIndexSlice", "_read_index_slice"),
        ("_initSortedSlice", "_init_sorted_slice"),
        ("_g_readSortedSlice", "_g_read_sorted_slice"),
        ("_readSortedSlice", "_read_sorted_slice"),
        ("getLRUbounds", "get_lru_bounds"),
        ("getLRUsorted", "get_lru_sorted"),
        ("_searchBinNA_b", "_search_bin_na_b"),
        ("_searchBinNA_ub", "_search_bin_na_ub"),
        ("_searchBinNA_s", "_search_bin_na_s"),
        ("_searchBinNA_us", "_search_bin_na_us"),
        ("_searchBinNA_i", "_search_bin_na_i"),
        ("_searchBinNA_ui", "_search_bin_na_ui"),
        ("_searchBinNA_ll", "_search_bin_na_ll"),
        ("_searchBinNA_ull", "_search_bin_na_ull"),
        ("_searchBinNA_e", "_search_bin_na_e"),
        ("_searchBinNA_f", "_search_bin_na_f"),
        ("_searchBinNA_d", "_search_bin_na_d"),
        ("_searchBinNA_g", "_search_bin_na_g"),
        # from leaf.py
        # ('parentNode', 'parentnode'),                       # kwarg
        ("objectID", "object_id"),  # property
        ("_processRangeRead", "_process_range_read"),
        ("_pointSelection", "_point_selection"),
        ("isVisible", "isvisible"),
        ("getAttr", "get_attr"),
        ("setAttr", "set_attr"),
        ("delAttr", "del_attr"),
        # from link.py
        # ('parentNode', 'parentnode'),                       # kwarg
        ("_g_getLinkClass", "_g_get_link_class"),
        # from linkextension
        ("linkExtension", "linkextension"),
        ("_getLinkClass", "_get_link_class"),
        ("_g_createHardLink", "_g_create_hard_link"),
        # from lrucacheextension
        ("lrucacheExtension", "lrucacheextension"),
        # from misc/enum.py
        ("_checkAndSetPair", "_check_and_set_pair"),
        ("_getContainer", "_get_container"),
        # from misc/proxydict.py
        ("containerRef", "containerref"),  # attr
        # from node.py
        # ('parentNode', 'parentnode'),                       # kwarg
        ("_g_logCreate", "_g_log_create"),
        ("_g_preKillHook", "_g_pre_kill_hook"),
        ("_g_checkOpen", "_g_check_open"),
        ("_g_setLocation", "_g_set_location"),
        ("_g_updateLocation", "_g_update_location"),
        ("_g_delLocation", "_g_del_location"),
        ("_g_updateDependent", "_g_update_dependent"),
        ("_g_removeAndLog", "_g_remove_and_log"),
        ("_g_logMove", "_g_log_move"),
        ("oldPathname", "oldpathname"),  # ??
        ("_g_copyAsChild", "_g_copy_as_child"),
        ("_f_isVisible", "_f_isvisible"),
        ("_g_checkGroup", "_g_check_group"),
        ("_g_checkNotContains", "_g_check_not_contains"),
        ("_g_maybeRemove", "_g_maybe_remove"),
        ("_f_getAttr", "_f_getattr"),
        ("_f_setAttr", "_f_setattr"),
        ("_f_delAttr", "_f_delattr"),
        ("_v_maxTreeDepth", "_v_maxtreedepth"),  # attr
        # from nodes/filenode.py
        ("newNode", "new_node"),
        ("openNode", "open_node"),
        ("_lineChunkSize", "_line_chunksize"),  # attr (private)
        ("_lineSeparator", "_line_separator"),  # attr (private)
        # ('getLineSeparator', 'get_line_separator'),        # dropped
        # ('setLineSeparator', 'set_line_separator'),        # dropped
        # ('delLineSeparator', 'del_line_separator'),        # dropped
        # ('lineSeparator', 'line_separator'),                # property -- dropped
        ("_notReadableError", "_not_readable_error"),
        ("_appendZeros", "_append_zeros"),
        ("getAttrs", "_get_attrs"),
        ("setAttrs", "_set_attrs"),
        ("delAttrs", "_del_attrs"),
        ("_setAttributes", "_set_attributes"),
        ("_checkAttributes", "_check_attributes"),
        ("_checkNotClosed", "_check_not_closed"),
        ("__allowedInitKwArgs", "__allowed_init_kwargs"),  # attr (private)
        ("_byteShape", "_byte_shape"),  # attr (private)
        ("_sizeToShape", "_size_to_shape"),  # attr (private)
        ("_vType", "_vtype"),  # attr (private)
        ("_vShape", "_vshape"),  # attr (private)
        # from path.py
        ("parentPath", "parentpath"),  # kwarg
        ("_pythonIdRE", "_python_id_re"),  # attr (private)
        ("_reservedIdRE", "_reserved_id_re"),  # attr (private)
        ("_hiddenNameRE", "_hidden_name_re"),  # attr (private)
        ("_hiddenPathRE", "_hidden_path_re"),  # attr (private)
        ("checkNameValidity", "check_name_validity"),
        ("joinPath", "join_path"),
        ("splitPath", "split_path"),
        ("isVisibleName", "isvisiblename"),
        ("isVisiblePath", "isvisiblepath"),
        # from registry.py
        ("className", "classname"),  # kwarg
        ("classNameDict", "class_name_dict"),  # data
        ("classIdDict", "class_id_dict"),  # data
        ("getClassByName", "get_class_by_name"),
        # from scripts/ptdump.py
        ("dumpLeaf", "dump_leaf"),
        ("dumpGroup", "dump_group"),
        # from scripts/ptrepack.py
        ("newdstGroup", "newdst_group"),
        ("recreateIndexes", "recreate_indexes"),
        ("copyLeaf", "copy_leaf"),
        # from table.py
        # ('parentNode', 'parentnode'),                       # kwarg
        ("_nxTypeFromNPType", "_nxtype_from_nptype"),  # data (private)
        ("_npSizeType", "_npsizetype"),  # data (private)
        ("_indexNameOf", "_index_name_of"),
        ("_indexPathnameOf", "_index_pathname_of"),
        ("_indexPathnameOfColumn", "_index_pathname_of_column"),
        ("_indexNameOf_", "_index_name_of_"),
        ("_indexPathnameOf_", "_index_pathname_of_"),
        ("_indexPathnameOfColumn_", "_index_pathname_of_column_"),
        ("_table__setautoIndex", "_table__setautoindex"),
        ("_table__getautoIndex", "_table__getautoindex"),
        ("_table__autoIndex", "_table__autoindex"),  # data (private)
        ("_table__whereIndexed", "_table__where_indexed"),
        ("createIndexesTable", "create_indexes_table"),
        ("createIndexesDescr", "create_indexes_descr"),
        ("_column__createIndex", "_column__create_index"),
        ("_autoIndex", "_autoindex"),  # attr
        ("autoIndex", "autoindex"),  # attr
        ("_useIndex", "_use_index"),
        ("_whereCondition", "_where_condition"),  # attr (private)
        ("_conditionCache", "_condition_cache"),  # attr (private)
        # ('_exprvarsCache', '_exprvars_cache'),
        (
            "_enabledIndexingInQueries",
            "_enabled_indexing_in_queries",
        ),  # attr (private)
        ("_emptyArrayCache", "_empty_array_cache"),  # attr (private)
        ("_getTypeColNames", "_get_type_col_names"),
        ("_getEnumMap", "_get_enum_map"),
        ("_cacheDescriptionData", "_cache_description_data"),
        ("_getColumnInstance", "_get_column_instance"),
        ("_checkColumn", "_check_column"),
        ("_disableIndexingInQueries", "_disable_indexing_in_queries"),
        ("_enableIndexingInQueries", "_enable_indexing_in_queries"),
        # ('_requiredExprVars', '_required_expr_vars'),
        ("_getConditionKey", "_get_condition_key"),
        ("_compileCondition", "_compile_condition"),
        ("willQueryUseIndexing", "will_query_use_indexing"),
        ("readWhere", "read_where"),
        ("whereAppend", "append_where"),
        ("getWhereList", "get_where_list"),
        ("_check_sortby_CSI", "_check_sortby_csi"),
        ("_readCoordinates", "_read_coordinates"),
        ("readCoordinates", "read_coordinates"),
        ("_saveBufferedRows", "_save_buffered_rows"),
        ("modifyCoordinates", "modify_coordinates"),
        ("modifyRows", "modify_rows"),
        ("modifyColumn", "modify_column"),
        ("modifyColumns", "modify_columns"),
        ("flushRowsToIndex", "flush_rows_to_index"),
        ("_addRowsToIndex", "_add_rows_to_index"),
        ("removeRows", "remove_rows"),
        ("_setColumnIndexing", "_set_column_indexing"),
        ("_markColumnsAsDirty", "_mark_columns_as_dirty"),
        ("_reIndex", "_reindex"),
        ("_doReIndex", "_do_reindex"),
        ("reIndex", "reindex"),
        ("reIndexDirty", "reindex_dirty"),
        ("_g_copyRows", "_g_copy_rows"),
        ("_g_copyRows_optim", "_g_copy_rows_optim"),
        ("_g_propIndexes", "_g_prop_indexes"),
        ("_g_updateTableLocation", "_g_update_table_location"),
        ("_tableFile", "_table_file"),  # attr (private)
        ("_tablePath", "_table_path"),  # attr (private)
        ("createIndex", "create_index"),
        ("createCSIndex", "create_csindex"),
        ("removeIndex", "remove_index"),
        # from tableextension
        ("tableExtension", "tableextension"),
        ("getNestedFieldCache", "get_nested_field_cache"),
        ("getNestedType", "get_nested_type"),
        ("_createTable", "_create_table"),
        ("_getInfo", "_get_info"),
        ("indexChunk", "indexchunk"),  # attr
        ("indexValid", "indexvalid"),  # attr
        ("indexValues", "indexvalues"),  # attr
        ("bufcoordsData", "bufcoords_data"),  # attr
        ("indexValuesData", "index_values_data"),  # attr
        ("chunkmapData", "chunkmap_data"),  # attr
        ("indexValidData", "index_valid_data"),  # attr
        ("whereCond", "wherecond"),  # attr
        ("iterseqMaxElements", "iterseq_max_elements"),  # attr
        ("IObuf", "iobuf"),  # attr
        ("IObufcpy", "iobufcpy"),  # attr
        ("_convertTime64_", "_convert_time64_"),
        ("_convertTypes", "_convert_types"),
        ("_newBuffer", "_new_buffer"),
        ("__next__inKernel", "__next__inkernel"),
        ("_fillCol", "_fill_col"),
        ("_flushBufferedRows", "_flush_buffered_rows"),
        ("_getUnsavedNrows", "_get_unsaved_nrows"),
        ("_flushModRows", "_flush_mod_rows"),
        # from undoredo.py
        ("moveToShadow", "move_to_shadow"),
        ("moveFromShadow", "move_from_shadow"),
        ("undoCreate", "undo_create"),
        ("redoCreate", "redo_create"),
        ("undoRemove", "undo_remove"),
        ("redoRemove", "redo_remove"),
        ("undoMove", "undo_move"),
        ("redoMove", "redo_move"),
        ("attrToShadow", "attr_to_shadow"),
        ("attrFromShadow", "attr_from_shadow"),
        ("undoAddAttr", "undo_add_attr"),
        ("redoAddAttr", "redo_add_attr"),
        ("undoDelAttr", "undo_del_attr"),
        ("redoDelAttr", "redo_del_attr"),
        # from utils.py
        ("convertToNPAtom", "convert_to_np_atom"),
        ("convertToNPAtom2", "convert_to_np_atom2"),
        ("checkFileAccess", "check_file_access"),
        ("logInstanceCreation", "log_instance_creation"),
        ("fetchLoggedInstances", "fetch_logged_instances"),
        ("countLoggedInstances", "count_logged_instances"),
        ("listLoggedInstances", "list_logged_instances"),
        ("dumpLoggedInstances", "dump_logged_instances"),
        ("detectNumberOfCores", "detect_number_of_cores"),
        # from utilsextension
        ("utilsExtension", "utilsextension"),
        ("PTTypeToHDF5", "pttype_to_hdf5"),  # data
        ("PTSpecialKinds", "pt_special_kinds"),  # data
        ("NPExtPrefixesToPTKinds", "npext_prefixes_to_ptkinds"),  # data
        ("HDF5ClassToString", "hdf5_class_to_string"),  # data
        ("setBloscMaxThreads", "set_blosc_max_threads"),
        ("silenceHDF5Messages", "silence_hdf5_messages"),
        ("isHDF5File", "is_hdf5_file"),
        ("isPyTablesFile", "is_pytables_file"),
        ("getHDF5Version", "get_hdf5_version"),
        ("getPyTablesVersion", "get_pytables_version"),
        ("whichLibVersion", "which_lib_version"),
        ("whichClass", "which_class"),
        ("getNestedField", "get_nested_field"),
        ("getFilters", "get_filters"),
        ("getTypeEnum", "get_type_enum"),
        ("enumFromHDF5", "enum_from_hdf5"),
        ("enumToHDF5", "enum_to_hdf5"),
        ("AtomToHDF5Type", "atom_to_hdf5_type"),
        ("loadEnum", "load_enum"),
        ("HDF5ToNPNestedType", "hdf5_to_np_nested_type"),
        ("HDF5ToNPExtType", "hdf5_to_np_ext_type"),
        ("AtomFromHDF5Type", "atom_from_hdf5_type"),
        ("createNestedType", "create_nested_type"),
        # from unimlemented.py
        ("_openUnImplemented", "_open_unimplemented"),
        # from vlarray.py
        # ('parentNode', 'parentnode'),                       # kwarg
        # ('expectedsizeinMB', 'expected_mb'),                # --> expectedrows
        # ('_v_expectedsizeinMB', '_v_expected_mb'),          # --> expectedrows
    ]
)

new2oldnames = {v: k for k, v in old2newnames.items()}

# Note that it is tempting to use the ast module here, but then this
# breaks transforming cython files.  So instead we are going to do the
# dumb thing with replace.


def make_subs(ns):
    """Make stubs."""
    names = new2oldnames if ns.reverse else old2newnames
    s = r"(?<=\W)({})(?=\W)".format("|".join(list(names)))
    if ns.ignore_previous:
        s += r"(?!\s*?=\s*?previous_api(_property)?\()"
        s += r"(?!\* to \*\w+\*)"
        s += r"(?!\* parameter has been renamed into \*\w+\*\.)"
        s += r"(?! is pending deprecation, import \w+ instead\.)"
    subs = re.compile(s, flags=re.MULTILINE)

    def repl(m):
        return names.get(m.group(1), m.group(0))

    return subs, repl


def main():
    """Implement the main CLI interface."""
    desc = (
        "PyTables 2.x -> 3.x API transition tool\n\n"
        "This tool displays to standard out, so it is \n"
        "common to pipe this to another file:\n\n"
        "$ pt2to3 oldfile.py > newfile.py"
    )
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        "-r",
        "--reverse",
        action="store_true",
        default=False,
        dest="reverse",
        help="reverts changes, going from 3.x -> 2.x.",
    )
    parser.add_argument(
        "-p",
        "--no-ignore-previous",
        action="store_false",
        default=True,
        dest="ignore_previous",
        help="ignores previous_api() calls.",
    )
    parser.add_argument(
        "-o", default=None, dest="output", help="output file to write to."
    )
    parser.add_argument(
        "-i",
        "--inplace",
        action="store_true",
        default=False,
        dest="inplace",
        help="overwrites the file in-place.",
    )
    parser.add_argument("filename", help="path to input file.")
    ns = parser.parse_args()

    if not Path(ns.filename).is_file():
        sys.exit(f"file {ns.filename!r} not found")
    src = Path(ns.filename).read_text()

    subs, repl = make_subs(ns)
    targ = subs.sub(repl, src)

    ns.output = ns.filename if ns.inplace else ns.output
    if ns.output is None:
        sys.stdout.write(targ)
    else:
        Path(ns.output).write_text(targ)


if __name__ == "__main__":
    main()



"""Required versions for PyTables dependencies."""

from packaging.version import Version

# **********************************************************************
#   Runtime requirements, keep versions in sync with
#   "Prerequisites" in the User's Guide.
# **********************************************************************

# Minimum recommended versions for mandatory packages
min_numpy_version = Version("1.20.0")
min_numexpr_version = Version("2.6.2")
# These are library versions, not the python modules
min_hdf5_version = Version("1.10.5")
min_blosc_version = Version("1.11.1")
min_blosc2_version = Version("2.11.0")



"""Functionality related with node paths in a PyTables file.

Variables
=========

`__docformat`__
    The format of documentation strings in this module.

"""

import re
import keyword
import warnings

from .exceptions import NaturalNameWarning

__docformat__ = "reStructuredText"
"""The format of documentation strings in this module."""


_python_id_re = re.compile("^[a-zA-Z_][a-zA-Z0-9_]*$")
"""Python identifier regular expression."""

_reserved_id_re = re.compile("^_[cfgv]_")
"""PyTables reserved identifier regular expression.

- c: class variables
- f: class public methods
- g: class private methods
- v: instance variables
"""

_hidden_name_re = re.compile("^_[pi]_")
"""Nodes with a name *matching* this expression are considered hidden.

For instance, ``name`` would be visible while ``_i_name`` would not.
"""

_hidden_path_re = re.compile("/_[pi]_")
"""Nodes with a path *containing* this expression are considered hidden.

For instance, a node with a pathname like ``/a/b/c`` would be visible
while nodes with pathnames like ``/a/c/_i_x`` or ``/a/_p_x/y`` would
not.
"""

_warn_info = (
    "you will not be able to use natural naming to access this object; "
    "using ``getattr()`` will still work, though"
)
"""Warning printed when a name will not be reachable through natural naming"""


def check_attribute_name(name: str) -> None:
    """Check the validity of the `name` of an attribute in AttributeSet.

    If the name is not valid, a ``ValueError`` is raised.  If it is
    valid but it can not be used with natural naming, a
    `NaturalNameWarning` is issued.

    >>> warnings.simplefilter("ignore")
    >>> check_attribute_name('a')
    >>> check_attribute_name('a_b')
    >>> check_attribute_name('a:b')         # NaturalNameWarning
    >>> check_attribute_name('/a/b')        # NaturalNameWarning
    >>> check_attribute_name('/')           # NaturalNameWarning
    >>> check_attribute_name('.')           # NaturalNameWarning
    >>> check_attribute_name('__members__')
    Traceback (most recent call last):
     ...
    ValueError: ``__members__`` is not allowed as an object name
    >>> check_attribute_name(1)
    Traceback (most recent call last):
     ...
    TypeError: object name is not a string: 1
    >>> check_attribute_name('')
    Traceback (most recent call last):
     ...
    ValueError: the empty string is not allowed as an object name
    """
    if not isinstance(name, str):  # Python >= 2.3
        raise TypeError(f"object name is not a string: {name!r}")

    if name == "":
        raise ValueError("the empty string is not allowed as an object name")

    # Check whether `name` is a valid Python identifier.
    if not _python_id_re.match(name):
        warnings.warn(
            "object name is not a valid Python identifier: %r; "
            "it does not match the pattern ``%s``; %s"
            % (name, _python_id_re.pattern, _warn_info),
            NaturalNameWarning,
            stacklevel=2,
        )
        return

    # However, Python identifiers and keywords have the same form.
    if keyword.iskeyword(name):
        warnings.warn(
            f"object name is a Python keyword: {name!r}; {_warn_info}",
            NaturalNameWarning,
            stacklevel=2,
        )
        return

    # Still, names starting with reserved prefixes are not allowed.
    if _reserved_id_re.match(name):
        raise ValueError(
            "object name starts with a reserved prefix: %r; "
            "it matches the pattern ``%s``" % (name, _reserved_id_re.pattern)
        )

    # ``__members__`` is the only exception to that rule.
    if name == "__members__":
        raise ValueError("``__members__`` is not allowed as an object name")


def check_name_validity(name: str) -> None:
    """Check the validity of the `name` of a Node object.

    Validity of Node names is more limited than attribute names.

    If the name is not valid, a ``ValueError`` is raised.  If it is
    valid but it can not be used with natural naming, a
    `NaturalNameWarning` is issued.

    >>> warnings.simplefilter("ignore")
    >>> check_name_validity('a')
    >>> check_name_validity('a_b')
    >>> check_name_validity('a:b')          # NaturalNameWarning
    >>> check_name_validity('/a/b')
    Traceback (most recent call last):
     ...
    ValueError: the ``/`` character is not allowed in object names: '/a/b'
    >>> check_name_validity('.')
    Traceback (most recent call last):
     ...
    ValueError: ``.`` is not allowed as an object name
    >>> check_name_validity('')
    Traceback (most recent call last):
     ...
    ValueError: the empty string is not allowed as an object name

    """
    check_attribute_name(name)

    # Check whether `name` is a valid HDF5 name.
    # http://hdfgroup.org/HDF5/doc/UG/03_Model.html#Structure
    if name == ".":
        raise ValueError("``.`` is not allowed as an object name")
    elif "/" in name:
        raise ValueError(
            "the ``/`` character is not allowed " "in object names: %r" % name
        )


def join_path(parentpath: str, name: str) -> str:
    """Join a *canonical* `parentpath` with a *non-empty* `name`.

    .. versionchanged:: 3.0
       The *parentPath* parameter has been renamed into *parentpath*.

    >>> join_path('/', 'foo')
    '/foo'
    >>> join_path('/foo', 'bar')
    '/foo/bar'
    >>> join_path('/foo', '/foo2/bar')
    '/foo/foo2/bar'
    >>> join_path('/foo', '/')
    '/foo'

    """
    if name.startswith("./"):  # Support relative paths (mainly for links)
        name = name[2:]
    if parentpath == "/" and name.startswith("/"):
        pstr = "%s" % name
    elif parentpath == "/" or name.startswith("/"):
        pstr = f"{parentpath}{name}"
    else:
        pstr = f"{parentpath}/{name}"
    if pstr.endswith("/"):
        pstr = pstr[:-1]
    return pstr


def split_path(path: str) -> (str, str):
    """Split a *canonical* `path` into a parent path and a node name.

    The result is returned as a tuple.  The parent path does not
    include a trailing slash.

    >>> split_path('/')
    ('/', '')
    >>> split_path('/foo/bar')
    ('/foo', 'bar')

    """
    lastslash = path.rfind("/")
    ppath = path[:lastslash]
    name = path[lastslash + 1 :]

    if ppath == "":
        ppath = "/"

    return (ppath, name)


def isvisiblename(name: str) -> bool:
    """Return `True` if  `name` makes the named node visible."""
    return _hidden_name_re.match(name) is None


def isvisiblepath(path: str) -> bool:
    """Return `True` if `path` makes the named node visible."""
    return _hidden_path_re.search(path) is None


def _test() -> None:
    """Run ``doctest`` on this module."""
    import doctest

    doctest.testmod()


if __name__ == "__main__":
    _test()



"""Utility functions."""

from __future__ import annotations

import os
import sys
import math
import weakref
import warnings
from time import perf_counter as clock
from typing import Any, Literal, TextIO, TYPE_CHECKING
from pathlib import Path
from collections.abc import Callable

import numpy as np
import numpy.typing as npt

from .flavor import array_of_flavor

if TYPE_CHECKING:
    from .atom import Atom

# The map between byteorders in NumPy and PyTables
byteorders = {
    ">": "big",
    "<": "little",
    "=": sys.byteorder,
    "|": "irrelevant",
}

# The type used for size values: indexes, coordinates, dimension
# lengths, row numbers, shapes, chunk shapes, byte counts...
SizeType = np.int64

copy_if_needed: bool | None

if np.lib.NumpyVersion(np.__version__) >= "2.0.0":
    copy_if_needed = None
elif np.lib.NumpyVersion(np.__version__) < "1.28.0":
    copy_if_needed = False
else:
    # 2.0.0 dev versions, handle cases where copy may or may not exist
    try:
        np.array([1]).__array__(copy=None)  # type: ignore[call-overload]
        copy_if_needed = None
    except TypeError:
        copy_if_needed = False


def correct_byteorder(ptype: str, byteorder: str) -> str:
    """Fix the byteorder depending on the PyTables types."""
    if ptype in ["string", "bool", "int8", "uint8", "object"]:
        return "irrelevant"
    else:
        return byteorder


def is_idx(index: Any) -> bool:
    """Check if an object can work as an index or not."""
    if type(index) is int:
        return True
    elif hasattr(index, "__index__"):
        # Exclude the array([idx]) as working as an index.  Fixes #303.
        if hasattr(index, "shape") and index.shape != ():
            return False
        try:
            index.__index__()
            if isinstance(index, bool):
                warnings.warn(
                    "using a boolean instead of an integer will result in an "
                    "error in the future",
                    DeprecationWarning,
                    stacklevel=2,
                )
            return True
        except TypeError:
            return False
    elif isinstance(index, np.integer):
        return True
    # For Python 2.4 one should test 0-dim and 1-dim, 1-elem arrays as well
    elif (
        isinstance(index, np.ndarray)
        and (index.shape == ())
        and index.dtype.str[1] == "i"
    ):
        return True

    return False


def idx2long(index: int | float | np.ndarray) -> int:
    """Convert a possible index into a long int."""
    try:
        if hasattr(index, "item"):
            return index.item()
        else:
            return int(index)
    except Exception:
        raise TypeError("not an integer type.")


# This is used in VLArray and EArray to produce NumPy object compliant
# with atom from a generic python type.  If copy is stated as True, it
# is assured that it will return a copy of the object and never the same
# object or a new one sharing the same memory.
def convert_to_np_atom(
    arr: npt.ArrayLike, atom: Atom, copy: bool | None = copy_if_needed
) -> np.ndarray:
    """Convert a generic object into a NumPy object compliant with atom."""
    # First, convert the object into a NumPy array
    nparr = array_of_flavor(arr, "numpy")
    # Copy of data if necessary for getting a contiguous buffer, or if
    # dtype is not the correct one.
    if atom.shape == ():
        # Scalar atom case
        nparr = np.array(nparr, dtype=atom.dtype, copy=copy)
    else:
        # Multidimensional atom case.  Addresses #133.
        # We need to use this strange way to obtain a dtype compliant
        # array because NumPy doesn't honor the shape of the dtype when
        # it is multidimensional.  See:
        # http://scipy.org/scipy/numpy/ticket/926
        # for details.
        # All of this is done just to taking advantage of the NumPy
        # broadcasting rules.
        newshape = nparr.shape[: -len(atom.dtype.shape)]
        nparr2 = np.empty(newshape, dtype=[("", atom.dtype)])
        nparr2["f0"][:] = nparr
        # Return a view (i.e. get rid of the record type)
        nparr = nparr2.view(atom.dtype)
    return nparr


# The next is used in Array, EArray and VLArray, and it is a bit more
# high level than convert_to_np_atom
def convert_to_np_atom2(obj: npt.ArrayLike, atom: Atom) -> np.ndarray:
    """Convert a generic object into a NumPy object compliant with atom."""
    # Check whether the object needs to be copied to make the operation
    # safe to in-place conversion.
    copy = True if atom.type in ["time64"] else copy_if_needed
    nparr = convert_to_np_atom(obj, atom, copy)
    # Finally, check the byteorder and change it if needed
    byteorder = byteorders[nparr.dtype.byteorder]
    if byteorder in ["little", "big"] and byteorder != sys.byteorder:
        # The byteorder needs to be fixed (a copy is made
        # so that the original array is not modified)
        nparr = nparr.byteswap()

    return nparr


def check_file_access(
    filename: str, mode: Literal["r", "w", "a", "r+"] = "r"
) -> None:
    """Check for file access in the specified `mode`.

    `mode` is one of the modes supported by `File` objects.  If the file
    indicated by `filename` can be accessed using that `mode`, the
    function ends successfully.  Else, an ``IOError`` is raised
    explaining the reason of the failure.

    All this paraphernalia is used to avoid the lengthy and scaring HDF5
    messages produced when there are problems opening a file.  No
    changes are ever made to the file system.

    """
    path = Path(filename).resolve()

    if mode == "r":
        # The file should be readable.
        if not os.access(path, os.F_OK):
            raise FileNotFoundError(f"``{path}`` does not exist")
        if not path.is_file():
            raise IsADirectoryError(f"``{path}`` is not a regular file")
        if not os.access(path, os.R_OK):
            raise PermissionError(
                f"file ``{path}`` exists but it can not be read"
            )
    elif mode == "w":
        if os.access(path, os.F_OK):
            # Since the file is not removed but replaced,
            # it must already be accessible to read and write operations.
            check_file_access(path, "r+")
        else:
            # A new file is going to be created,
            # so the directory should be writable.
            if not os.access(path.parent, os.F_OK):
                raise FileNotFoundError(f"``{path.parent}`` does not exist")
            if not path.parent.is_dir():
                raise NotADirectoryError(
                    f"``{path.parent}`` is not a directory"
                )
            if not os.access(path.parent, os.W_OK):
                raise PermissionError(
                    f"directory ``{path.parent}`` exists but it can not be "
                    f"written"
                )
    elif mode == "a":
        if os.access(path, os.F_OK):
            check_file_access(path, "r+")
        else:
            check_file_access(path, "w")
    elif mode == "r+":
        check_file_access(path, "r")
        if not os.access(path, os.W_OK):
            raise PermissionError(
                f"file ``{path}`` exists but it can not be written"
            )
    else:
        raise ValueError(f"invalid mode: {mode!r}")


def lazyattr(fget: Callable[[Any], Any]) -> property:
    """Create a *lazy attribute* from the result of `fget`.

    This function is intended to be used as a *method decorator*.  It
    returns a *property* which caches the result of calling the `fget`
    instance method.  The docstring of `fget` is used for the property
    itself.  For instance:

    >>> class MyClass(object):
    ...     @lazyattr
    ...     def attribute(self):
    ...         'Attribute description.'
    ...         print('creating value')
    ...         return 10
    ...
    >>> type(MyClass.attribute)
    <class 'property'>
    >>> MyClass.attribute.__doc__
    'Attribute description.'
    >>> obj = MyClass()
    >>> obj.__dict__
    {}
    >>> obj.attribute
    creating value
    10
    >>> obj.__dict__
    {'attribute': 10}
    >>> obj.attribute
    10
    >>> del obj.attribute
    Traceback (most recent call last):
      ...
    AttributeError: ...

    .. warning::

        Please note that this decorator *changes the type of the
        decorated object* from an instance method into a property.

    """
    name = fget.__name__

    def newfget(self):
        mydict = self.__dict__
        if name in mydict:
            return mydict[name]
        mydict[name] = value = fget(self)
        return value

    return property(newfget, None, None, fget.__doc__)


def show_stats(explain: str, tref: float, encoding=None) -> float:
    """Show the used memory (only works for Linux 2.6.x)."""
    for line in Path("/proc/self/status").read_text().splitlines():
        if line.startswith("VmSize:"):
            vmsize = int(line.split()[1])
        elif line.startswith("VmRSS:"):
            vmrss = int(line.split()[1])
        elif line.startswith("VmData:"):
            vmdata = int(line.split()[1])
        elif line.startswith("VmStk:"):
            vmstk = int(line.split()[1])
        elif line.startswith("VmExe:"):
            vmexe = int(line.split()[1])
        elif line.startswith("VmLib:"):
            vmlib = int(line.split()[1])
    print("Memory usage: ******* %s *******" % explain)
    print(f"VmSize: {vmsize:>7} kB\tVmRSS: {vmrss:>7} kB")
    print(f"VmData: {vmdata:>7} kB\tVmStk: {vmstk:>7} kB")
    print(f"VmExe:  {vmexe:>7} kB\tVmLib: {vmlib:>7} kB")
    tnow = clock()
    print(f"WallClock time: {tnow - tref:.3f}")
    return tnow


# truncate data before calling __setitem__, to improve compression ratio
# this function is taken verbatim from netcdf4-python
def quantize(data: npt.ArrayLike, least_significant_digit: int):
    """Quantize data to improve compression.

    Data is quantized using around(scale*data)/scale, where scale is
    2**bits, and bits is determined from the least_significant_digit.

    For example, if least_significant_digit=1, bits will be 4.

    """
    exp = -least_significant_digit
    exp = math.floor(exp) if exp < 0 else math.ceil(exp)
    bits = math.ceil(math.log2(10**-exp))
    scale = 2**bits
    datout = np.around(scale * data) / scale

    return datout


# Utilities to detect leaked instances.  See recipe 14.10 of the Python
# Cookbook by Martelli & Ascher.
tracked_classes: dict[str, list[weakref.ReferenceType]] = {}


def log_instance_creation(instance: Any, name: str | None = None) -> None:
    """Log instance creation."""
    if name is None:
        name = instance.__class__.__name__
        if name not in tracked_classes:
            tracked_classes[name] = []
        tracked_classes[name].append(weakref.ref(instance))


def string_to_classes(s: str) -> list[str]:
    """Return the list of tracked classes matching the input string."""
    if s == "*":
        c = sorted(tracked_classes)
        return c
    else:
        return s.split()


def fetch_logged_instances(classes: str = "*") -> list[tuple[str, int]]:
    """Return the list of logged instances."""
    classnames = string_to_classes(classes)
    return [(cn, len(tracked_classes[cn])) for cn in classnames]


def count_logged_instances(classes: str, file: TextIO = sys.stdout) -> None:
    """Write to file the number of logged instances."""
    for classname in string_to_classes(classes):
        file.write(f"{classname}: {len(tracked_classes[classname])}\n")


def list_logged_instances(classes: str, file: TextIO = sys.stdout) -> None:
    """Write to file the list of loggen instances."""
    for classname in string_to_classes(classes):
        file.write(f"\n{classname}:\n")
        for ref in tracked_classes[classname]:
            obj = ref()
            if obj is not None:
                file.write("    %s\n" % repr(obj))


def dump_logged_instances(classes: str, file: TextIO = sys.stdout) -> None:
    """Dump the logged instances."""
    for classname in string_to_classes(classes):
        file.write(f"\n{classname}:\n")
        for ref in tracked_classes[classname]:
            obj = ref()
            if obj is not None:
                file.write("    %s:\n" % obj)
                for key, value in obj.__dict__.items():
                    file.write(f"        {key:>20} : {value}\n")


#
# A class useful for cache usage
#
class CacheDict(dict):
    """A dictionary that prevents itself from growing too much."""

    def __init__(self, maxentries: int) -> None:
        self.maxentries = maxentries
        super().__init__(self)

    def __setitem__(self, key: str, value: Any) -> None:
        # Protection against growing the cache too much
        if len(self) > self.maxentries:
            # Remove a 10% of (arbitrary) elements from the cache
            entries_to_remove = self.maxentries / 10
            for k in list(self)[:entries_to_remove]:
                super().__delitem__(k)
        super().__setitem__(key, value)


class NailedDict:
    """A dictionary which ignores its items when it has nails on it."""

    def __init__(self, maxentries: int) -> None:
        self.maxentries = maxentries
        self._cache: dict = {}
        self._nailcount = 0

    # Only a restricted set of dictionary methods are supported.  That
    # is why we buy instead of inherit.

    # The following are intended to be used by ``Table`` code changing
    # the set of usable indexes.

    def clear(self) -> None:
        """Clear teh dictionsry."""
        self._cache.clear()

    def nail(self) -> None:
        """Increase the nail count."""
        self._nailcount += 1

    def unnail(self) -> None:
        """Decrease the nail count."""
        self._nailcount -= 1

    # The following are intended to be used by ``Table`` code handling
    # conditions.

    def __contains__(self, key: Any) -> bool:
        if self._nailcount > 0:
            return False
        return key in self._cache

    def __getitem__(self, key: Any) -> Any:
        if self._nailcount > 0:
            raise KeyError(key)
        return self._cache[key]

    def get(self, key: Any, default: Any | None = None) -> Any:
        """Return the value for the specified key."""
        if self._nailcount > 0:
            return default
        return self._cache.get(key, default)

    def __setitem__(self, key: Any, value: Any) -> None:
        if self._nailcount > 0:
            return
        cache = self._cache
        # Protection against growing the cache too much
        if len(cache) > self.maxentries:
            # Remove a 10% of (arbitrary) elements from the cache
            entries_to_remove = max(self.maxentries // 10, 1)
            for k in list(cache)[:entries_to_remove]:
                del cache[k]
        cache[key] = value


def detect_number_of_cores() -> int:
    """Detect the number of cores on a system."""
    # Linux, Unix and MacOS:
    if hasattr(os, "sysconf"):
        if "SC_NPROCESSORS_ONLN" in os.sysconf_names:
            # Linux & Unix:
            ncpus = os.sysconf("SC_NPROCESSORS_ONLN")
            if isinstance(ncpus, int) and ncpus > 0:
                return ncpus
        else:  # OSX:
            return int(os.popen2("sysctl -n hw.ncpu")[1].read())
    # Windows:
    if "NUMBER_OF_PROCESSORS" in os.environ:
        ncpus = int(os.environ["NUMBER_OF_PROCESSORS"])
        if ncpus > 0:
            return ncpus
    return 1  # Default


def _test() -> None:
    """Run ``doctest`` on this module."""
    import doctest

    doctest.testmod()


if __name__ == "__main__":
    _test()



"""This utility lets you repack your data files in a flexible way.

Pass the flag -h to this for help on usage.

"""

import sys
import argparse
import warnings
from time import perf_counter as clock
from time import process_time as cpuclock
from pathlib import Path

import tables as tb

# Global variables
verbose = False
regoldindexes = True
createsysattrs = True

numpy_aliases = [
    "numeric",
    "Numeric",
    "numarray",
    "NumArray",
    "CharArray",
]


def newdst_group(dstfileh, dstgroup, title, filters):
    """Return a new destination group."""
    group = dstfileh.root
    # Now, create the new group. This works even if dstgroup == '/'
    for nodename in dstgroup.split("/"):
        if nodename == "":
            continue
        # First try if possible intermediate groups does already exist.
        try:
            group2 = dstfileh.get_node(group, nodename)
        except tb.exceptions.NoSuchNodeError:
            # The group does not exist. Create it.
            group2 = dstfileh.create_group(
                group, nodename, title=title, filters=filters
            )
        group = group2
    return group


def recreate_indexes(table, dstfileh, dsttable):
    """Re-create indexes."""
    listoldindexes = table._listoldindexes
    if listoldindexes != []:
        if not regoldindexes:
            if verbose:
                print(
                    f"[I]Not regenerating indexes for table: "
                    f"'{dstfileh.filename}:{dsttable._v_pathname}'"
                )
            return
        # Now, recreate the indexed columns
        if verbose:
            print(
                f"[I]Regenerating indexes for table: "
                f"'{dstfileh.filename}:{dsttable._v_pathname}'"
            )
        for colname in listoldindexes:
            if verbose:
                print("[I]Indexing column: '%s'. Please wait..." % colname)
            colobj = dsttable.cols._f_col(colname)
            # We don't specify the filters for the indexes
            colobj.create_index(filters=None)


def copy_leaf(
    srcfile,
    dstfile,
    srcnode,
    dstnode,
    title,
    filters,
    copyuserattrs,
    overwritefile,
    overwrtnodes,
    stats,
    start,
    stop,
    step,
    chunkshape,
    sortby,
    check_CSI,  # noqa: N803
    propindexes,
    upgradeflavors,
    allow_padding,
):
    """Copy a leaf node."""
    # Open the source file
    srcfileh = tb.open_file(srcfile, "r", allow_padding=allow_padding)
    # Get the source node (that should exist)
    srcnode = srcfileh.get_node(srcnode)

    # Get the destination node and its parent
    last_slash = dstnode.rindex("/")
    if last_slash == len(dstnode) - 1:
        # print("Detected a trailing slash in destination node. "
        #       "Interpreting it as a destination group.")
        dstgroup = dstnode[:-1]
    elif last_slash > 0:
        dstgroup = dstnode[:last_slash]
    else:
        dstgroup = "/"
    dstleaf = dstnode[last_slash + 1 :]
    if dstleaf == "":
        dstleaf = srcnode.name
    # Check whether the destination group exists or not
    if Path(dstfile).is_file() and not overwritefile:
        dstfileh = tb.open_file(
            dstfile,
            "a",
            pytables_sys_attrs=createsysattrs,
            allow_padding=allow_padding,
        )
        try:
            dstgroup = dstfileh.get_node(dstgroup)
        except Exception:
            # The dstgroup does not seem to exist. Try creating it.
            dstgroup = newdst_group(dstfileh, dstgroup, title, filters)
        else:
            # The node exists, but it is really a group?
            if not isinstance(dstgroup, tb.group.Group):
                # No. Should we overwrite it?
                if overwrtnodes:
                    parent = dstgroup._v_parent
                    last_slash = dstgroup._v_pathname.rindex("/")
                    dstgroupname = dstgroup._v_pathname[last_slash + 1 :]
                    dstgroup.remove()
                    dstgroup = dstfileh.create_group(
                        parent, dstgroupname, title=title, filters=filters
                    )
                else:
                    raise RuntimeError(
                        "Please check that the node names are "
                        "not duplicated in destination, and "
                        "if so, add the --overwrite-nodes "
                        "flag if desired."
                    )
    else:
        # The destination file does not exist or will be overwritten.
        dstfileh = tb.open_file(
            dstfile,
            "w",
            title=title,
            filters=filters,
            pytables_sys_attrs=createsysattrs,
            allow_padding=allow_padding,
        )
        dstgroup = newdst_group(dstfileh, dstgroup, title="", filters=filters)

    # Finally, copy srcnode to dstnode
    try:
        dstnode = srcnode.copy(
            dstgroup,
            dstleaf,
            filters=filters,
            copyuserattrs=copyuserattrs,
            overwrite=overwrtnodes,
            stats=stats,
            start=start,
            stop=stop,
            step=step,
            chunkshape=chunkshape,
            sortby=sortby,
            check_CSI=check_CSI,
            propindexes=propindexes,
        )
    except Exception:
        type_, value, traceback = sys.exc_info()
        print(
            f"Problems doing the copy from '{srcfile}:{srcnode}' to "
            f"'{dstfile}:{dstnode}'"
        )
        print(f"The error was --> {type_}: {value}")
        print("The destination file looks like:\n", dstfileh)
        # Close all the open files:
        srcfileh.close()
        dstfileh.close()
        raise RuntimeError(
            "Please check that the node names are not "
            "duplicated in destination, and if so, add "
            "the --overwrite-nodes flag if desired."
        )

    # Upgrade flavors in dstnode, if required
    if upgradeflavors:
        if srcfileh.format_version.startswith("1"):
            # Remove original flavor in case the source file has 1.x format
            dstnode.del_attr("FLAVOR")
        elif srcfileh.format_version < "2.1":
            if dstnode.get_attr("FLAVOR") in numpy_aliases:
                dstnode.set_attr("FLAVOR", tb.flavor.internal_flavor)

    # Recreate possible old indexes in destination node
    if srcnode._c_classid == "TABLE":
        recreate_indexes(srcnode, dstfileh, dstnode)

    # Close all the open files:
    srcfileh.close()
    dstfileh.close()


def copy_children(
    srcfile,
    dstfile,
    srcgroup,
    dstgroup,
    title,
    recursive,
    filters,
    copyuserattrs,
    overwritefile,
    overwrtnodes,
    stats,
    start,
    stop,
    step,
    chunkshape,
    sortby,
    check_CSI,  # noqa: N803
    propindexes,
    upgradeflavors,
    allow_padding,
    use_hardlinks=True,
):
    """Copy the children from source group to destination group."""
    # Open the source file with srcgroup as root_uep
    srcfileh = tb.open_file(
        srcfile, "r", root_uep=srcgroup, allow_padding=allow_padding
    )
    #  Assign the root to srcgroup
    srcgroup = srcfileh.root

    created_dstgroup = False
    # Check whether the destination group exists or not
    if Path(dstfile).is_file() and not overwritefile:
        dstfileh = tb.open_file(
            dstfile,
            "a",
            pytables_sys_attrs=createsysattrs,
            allow_padding=allow_padding,
        )
        try:
            dstgroup = dstfileh.get_node(dstgroup)
        except tb.exceptions.NoSuchNodeError:
            # The dstgroup does not seem to exist. Try creating it.
            dstgroup = newdst_group(dstfileh, dstgroup, title, filters)
            created_dstgroup = True
        else:
            # The node exists, but it is really a group?
            if not isinstance(dstgroup, tb.group.Group):
                # No. Should we overwrite it?
                if overwrtnodes:
                    parent = dstgroup._v_parent
                    last_slash = dstgroup._v_pathname.rindex("/")
                    dstgroupname = dstgroup._v_pathname[last_slash + 1 :]
                    dstgroup.remove()
                    dstgroup = dstfileh.create_group(
                        parent, dstgroupname, title=title, filters=filters
                    )
                else:
                    raise RuntimeError(
                        "Please check that the node names are "
                        "not duplicated in destination, and "
                        "if so, add the --overwrite-nodes "
                        "flag if desired."
                    )
    else:
        # The destination file does not exist or will be overwritten.
        dstfileh = tb.open_file(
            dstfile,
            "w",
            title=title,
            filters=filters,
            pytables_sys_attrs=createsysattrs,
            allow_padding=allow_padding,
        )
        dstgroup = newdst_group(dstfileh, dstgroup, title="", filters=filters)
        created_dstgroup = True

    # Copy the attributes to dstgroup, if needed
    if created_dstgroup and copyuserattrs:
        srcgroup._v_attrs._f_copy(dstgroup)

    # Finally, copy srcgroup children to dstgroup
    try:
        srcgroup._f_copy_children(
            dstgroup,
            recursive=recursive,
            filters=filters,
            copyuserattrs=copyuserattrs,
            overwrite=overwrtnodes,
            stats=stats,
            start=start,
            stop=stop,
            step=step,
            chunkshape=chunkshape,
            sortby=sortby,
            check_CSI=check_CSI,
            propindexes=propindexes,
            use_hardlinks=use_hardlinks,
        )
    except Exception:
        type_, value, traceback = sys.exc_info()
        print(
            "Problems doing the copy from '%s:%s' to '%s:%s'"
            % (srcfile, srcgroup, dstfile, dstgroup)
        )
        print(f"The error was --> {type_}: {value}")
        print("The destination file looks like:\n", dstfileh)
        # Close all the open files:
        srcfileh.close()
        dstfileh.close()
        raise RuntimeError(
            "Please check that the node names are not "
            "duplicated in destination, and if so, add the "
            "--overwrite-nodes flag if desired. In "
            "particular, pay attention that root_uep is not "
            "fooling you."
        )

    # Upgrade flavors in dstnode, if required
    if upgradeflavors:
        for dstnode in dstgroup._f_walknodes("Leaf"):
            if srcfileh.format_version.startswith("1"):
                # Remove original flavor in case the source file has 1.x format
                dstnode.del_attr("FLAVOR")
            elif srcfileh.format_version < "2.1":
                if dstnode.get_attr("FLAVOR") in numpy_aliases:
                    dstnode.set_attr("FLAVOR", tb.flavor.internal_flavor)

    # Convert the remaining tables with old indexes (if any)
    for table in srcgroup._f_walknodes("Table"):
        dsttable = dstfileh.get_node(dstgroup, table._v_pathname)
        recreate_indexes(table, dstfileh, dsttable)

    # Close all the open files:
    srcfileh.close()
    dstfileh.close()


def _get_parser():
    parser = argparse.ArgumentParser(
        description=(
            "This utility is very powerful and lets you copy any "
            "leaf, group or complete subtree into another file. "
            "During the copy process you are allowed to change the filter "
            "properties if you want so. Also, in the case of duplicated "
            "pathnames, you can decide if you want to overwrite already "
            "existing nodes on the destination file. Generally speaking, "
            "ptrepack can be useful in may situations, like replicating a "
            "subtree in another file, change the filters in objects and see "
            "how affect this to the compression degree or I/O performance, "
            "consolidating specific data in repositories or even *importing* "
            "generic HDF5 files and create true PyTables counterparts."
        )
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="show verbose information",
    )
    parser.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        dest="overwritefile",
        help="overwrite destination file",
    )
    parser.add_argument(
        "-R",
        "--range",
        dest="rng",
        metavar="RANGE",
        help=(
            "select a RANGE of rows (in the form 'start,stop,step') "
            "during the copy of *all* the leaves. "
            "Default values are 'None, None, 1', which means a copy of all "
            "the rows."
        ),
    )
    parser.add_argument(
        "--non-recursive",
        action="store_false",
        default=True,
        dest="recursive",
        help="do not do a recursive copy. Default is to do it",
    )
    parser.add_argument(
        "--dest-title",
        dest="title",
        default="",
        help="title for the new file (if not specified, the source is copied)",
    )
    parser.add_argument(
        "--dont-create-sysattrs",
        action="store_false",
        default=True,
        dest="createsysattrs",
        help="do not create sys attrs (default is to do it)",
    )
    parser.add_argument(
        "--dont-copy-userattrs",
        action="store_false",
        default=True,
        dest="copyuserattrs",
        help="do not copy the user attrs (default is to do it)",
    )
    parser.add_argument(
        "--overwrite-nodes",
        action="store_true",
        dest="overwrtnodes",
        help=(
            "overwrite destination nodes if they exist. "
            "Default is to not overwrite them"
        ),
    )
    parser.add_argument(
        "--complevel",
        type=int,
        default=0,
        help=(
            "set a compression level (0 for no compression, which is the "
            "default)"
        ),
    )
    parser.add_argument(
        "--complib",
        choices=(
            "zlib",
            "lzo",
            "bzip2",
            "blosc",
            "blosc:blosclz",
            "blosc:lz4",
            "blosc:lz4hc",
            "blosc:zlib",
            "blosc:zstd",
            "blosc2",
            "blosc2:blosclz",
            "blosc2:lz4",
            "blosc2:lz4hc",
            "blosc2:zlib",
            "blosc2:zstd",
        ),
        default="zlib",
        help=(
            "set the compression library to be used during the copy. "
            "Defaults to %(default)s"
        ),
    )
    parser.add_argument(
        "--shuffle",
        type=int,
        choices=(0, 1),
        help=(
            "activate or not the shuffle filter (default is active if "
            "complevel > 0)"
        ),
    )
    parser.add_argument(
        "--bitshuffle",
        type=int,
        choices=(0, 1),
        help="activate or not the bitshuffle filter (not active by default)",
    )
    parser.add_argument(
        "--fletcher32",
        type=int,
        choices=(0, 1),
        help=(
            "whether to activate or not the fletcher32 filter (not active "
            "by default)"
        ),
    )
    parser.add_argument(
        "--keep-source-filters",
        action="store_true",
        dest="keepfilters",
        help=(
            "use the original filters in source files. "
            "The default is not doing that if any of --complevel, --complib, "
            "--shuffle --bitshuffle or --fletcher32 option is specified"
        ),
    )
    parser.add_argument(
        "--chunkshape",
        default="keep",
        help=(
            "set a chunkshape. "
            "Possible options are: 'keep' | 'auto' | int | tuple. "
            "A value of 'auto' computes a sensible value for the chunkshape "
            "of the leaves copied. "
            "The default is to 'keep' the original value"
        ),
    )
    parser.add_argument(
        "--upgrade-flavors",
        action="store_true",
        dest="upgradeflavors",
        help=(
            "when repacking PyTables 1.x or PyTables 2.x files, the flavor "
            "of leaves will be unset. With this, such a leaves will be "
            "serialized as objects with the internal flavor "
            "('numpy' for 3.x series)"
        ),
    )
    parser.add_argument(
        "--dont-regenerate-old-indexes",
        action="store_false",
        default=True,
        dest="regoldindexes",
        help=(
            "disable regenerating old indexes. "
            "The default is to regenerate old indexes as they are found"
        ),
    )
    parser.add_argument(
        "--sortby",
        metavar="COLUMN",
        help=(
            "do a table copy sorted by the index in 'column'. "
            "For reversing the order, use a negative value in the 'step' "
            "part of 'RANGE' (see '-r' flag).  Only applies to table objects"
        ),
    )
    parser.add_argument(
        "--checkCSI",
        action="store_true",
        help="force the check for a CSI index for the --sortby column",
    )
    parser.add_argument(
        "--propindexes",
        action="store_true",
        help=(
            "propagate the indexes existing in original tables. The default "
            "is to not propagate them.  Only applies to table objects"
        ),
    )
    parser.add_argument(
        "--dont-allow-padding",
        action="store_true",
        dest="dont_allow_padding",
        help=(
            "remove the possible padding in compound types in source files. "
            "The default is to propagate it.  Only applies to table objects"
        ),
    )
    parser.add_argument(
        "src",
        metavar="sourcefile:sourcegroup",
        help="source file/group",
    )
    parser.add_argument(
        "dst",
        metavar="destfile:destgroup",
        help="destination file/group",
    )

    return parser


def main():
    """Implement the main CLI interface."""
    global verbose
    global regoldindexes
    global createsysattrs

    parser = _get_parser()
    args = parser.parse_args()

    # check arguments
    if args.rng:
        try:
            args.rng = eval("slice(" + args.rng + ")")
        except Exception:
            parser.error("Error when getting the range parameter.")

    if args.chunkshape.isdigit() or args.chunkshape.startswith("("):
        args.chunkshape = eval(args.chunkshape)

    if args.complevel < 0 or args.complevel > 9:
        parser.error(
            'invalid "complevel" value, it sould be in te range [0, 9]'
        )

    # Catch the files passed as the last arguments
    src = args.src.rsplit(":", 1)
    dst = args.dst.rsplit(":", 1)
    if len(src) == 1:
        srcfile, srcnode = src[0], "/"
    else:
        srcfile, srcnode = src
    if len(dst) == 1:
        dstfile, dstnode = dst[0], "/"
    else:
        dstfile, dstnode = dst

    if srcnode == "":
        # case where filename == "filename:" instead of "filename:/"
        srcnode = "/"

    if dstnode == "":
        # case where filename == "filename:" instead of "filename:/"
        dstnode = "/"

    # Ignore the warnings for tables that contains oldindexes
    # (these will be handled by the copying routines)
    warnings.filterwarnings("ignore", category=tb.exceptions.OldIndexWarning)

    # Ignore the flavors warnings during upgrading flavor operations
    if args.upgradeflavors:
        warnings.filterwarnings("ignore", category=tb.exceptions.FlavorWarning)

    # Build the Filters instance
    filter_params = (
        args.complevel,
        args.complib,
        args.shuffle,
        args.bitshuffle,
        args.fletcher32,
    )
    if filter_params == (None,) * 4 or args.keepfilters:
        filters = None
    else:
        if args.complevel is None:
            args.complevel = 0
        if args.shuffle is None:
            if args.complevel > 0:
                args.shuffle = True
            else:
                args.shuffle = False
        if args.bitshuffle is None:
            args.bitshuffle = False
        if args.bitshuffle:
            # Shuffle and bitshuffle are mutually exclusive
            args.shuffle = False
        if args.complib is None:
            args.complib = "zlib"
        if args.fletcher32 is None:
            args.fletcher32 = False
        filters = tb.leaf.Filters(
            complevel=args.complevel,
            complib=args.complib,
            shuffle=args.shuffle,
            bitshuffle=args.bitshuffle,
            fletcher32=args.fletcher32,
        )

    # The start, stop and step params:
    start, stop, step = None, None, 1  # Defaults
    if args.rng:
        start, stop, step = args.rng.start, args.rng.stop, args.rng.step

    # Set globals
    verbose = args.verbose
    regoldindexes = args.regoldindexes
    createsysattrs = args.createsysattrs

    # Some timing
    t1 = clock()
    cpu1 = cpuclock()
    # Copy the file
    if verbose:
        print("+=+" * 20)
        print("Recursive copy:", args.recursive)
        print("Applying filters:", filters)
        if args.sortby is not None:
            print("Sorting table(s) by column:", args.sortby)
            print("Forcing a CSI creation:", args.checkCSI)
        if args.propindexes:
            print("Recreating indexes in copied table(s)")
        print(f"Start copying {srcfile}:{srcnode} to {dstfile}:{dstnode}")
        print("+=+" * 20)

    allow_padding = not args.dont_allow_padding
    # Check whether the specified source node is a group or a leaf
    h5srcfile = tb.open_file(srcfile, "r", allow_padding=allow_padding)
    srcnodeobject = h5srcfile.get_node(srcnode)

    # Close the file again
    h5srcfile.close()

    stats = {"groups": 0, "leaves": 0, "links": 0, "bytes": 0, "hardlinks": 0}
    if isinstance(srcnodeobject, tb.group.Group):
        copy_children(
            srcfile,
            dstfile,
            srcnode,
            dstnode,
            title=args.title,
            recursive=args.recursive,
            filters=filters,
            copyuserattrs=args.copyuserattrs,
            overwritefile=args.overwritefile,
            overwrtnodes=args.overwrtnodes,
            stats=stats,
            start=start,
            stop=stop,
            step=step,
            chunkshape=args.chunkshape,
            sortby=args.sortby,
            check_CSI=args.checkCSI,
            propindexes=args.propindexes,
            upgradeflavors=args.upgradeflavors,
            allow_padding=allow_padding,
            use_hardlinks=True,
        )
    else:
        # If not a Group, it should be a Leaf
        copy_leaf(
            srcfile,
            dstfile,
            srcnode,
            dstnode,
            title=args.title,
            filters=filters,
            copyuserattrs=args.copyuserattrs,
            overwritefile=args.overwritefile,
            overwrtnodes=args.overwrtnodes,
            stats=stats,
            start=start,
            stop=stop,
            step=step,
            chunkshape=args.chunkshape,
            sortby=args.sortby,
            check_CSI=args.checkCSI,
            propindexes=args.propindexes,
            upgradeflavors=args.upgradeflavors,
            allow_padding=allow_padding,
        )

    # Gather some statistics
    t2 = clock()
    cpu2 = cpuclock()
    tcopy = t2 - t1
    cpucopy = cpu2 - cpu1
    if verbose:
        ngroups = stats["groups"]
        nleaves = stats["leaves"]
        nlinks = stats["links"]
        nhardlinks = stats["hardlinks"]
        nbytescopied = stats["bytes"]
        nnodes = ngroups + nleaves + nlinks + nhardlinks

        print(
            "Groups copied:",
            ngroups,
            ", Leaves copied:",
            nleaves,
            ", Links copied:",
            nlinks,
            ", Hard links copied:",
            nhardlinks,
        )
        if args.copyuserattrs:
            print("User attrs copied")
        else:
            print("User attrs not copied")
        print(f"KBytes copied: {nbytescopied / 1024:.3f}")
        print(
            f"Time copying: {tcopy:.3f} s (real) {cpucopy:.3f} s "
            f"(cpu)  {cpucopy / tcopy:.0%}"
        )
        print(f"Copied nodes/sec: {nnodes / tcopy:.1f}")
        print(f"Copied KB/s : {nbytescopied / tcopy / 1024:.0f}")



"""PyTables, hierarchical datasets in Python.

:URL: http://www.pytables.org/

PyTables is a package for managing hierarchical datasets and designed
to efficiently cope with extremely large amounts of data.

"""


# Load the blosc2 library:
# 1. In tables.libs/ sibling (delvewheel, Windows-only)
# 2. In tables
# 3. In site-packages/blosc2/lib/ (venv, conda env, or system Python; same one where this tables is running)
# 4. Without a path (default, only the filename)
def _load_blosc2():
    import ctypes
    import platform
    import sysconfig
    from pathlib import Path

    search_paths = (
        # "delvewheel"
        Path(__file__).parent.with_suffix(".libs"),
        # tables package
        Path(__file__).parent,
        # "site-packages"
        Path(sysconfig.get_path("platlib")) / "blosc2" / "lib",
        # "site-packages" purelib - this should be redundant
        Path(sysconfig.get_path("purelib")) / "blosc2" / "lib",
        # "default"
        "",
    )
    platform_system = platform.system()
    ext = (
        "so"
        if platform_system == "Linux"
        else ("dylib" if platform_system == "Darwin" else "dll")
    )
    lib_name = "blosc2"
    lib_file = f"lib{lib_name}.{ext}"

    for where in search_paths:
        lib_path = Path(where) / lib_file
        if where == "" or lib_path.exists():
            try:
                ctypes.CDLL(str(lib_path))  # may be Path in Python 3.12+
                return True
            except OSError:
                pass

    import ctypes.util

    if ctypes.util.find_library(lib_name):
        return True

    return False


if not _load_blosc2():
    raise RuntimeError("Blosc2 library not found.")

from ._version import __version__

# Necessary imports to get versions stored on the cython extension
from .utilsextension import get_hdf5_version as _get_hdf5_version

hdf5_version = _get_hdf5_version()
"""The underlying HDF5 library version number.

.. versionadded:: 3.0

"""

from .atom import *
from .file import File, open_file, copy_file
from .leaf import Leaf, ChunkInfo
from .node import Node
from .array import Array
from .group import Group
from .table import Table, Cols, Column
from .tests import print_versions, test
from .carray import CArray
from .earray import EArray
from .flavor import restrict_flavors
from .filters import Filters
from .vlarray import VLArray
from .misc.enum import Enum

# Import the user classes from the proper modules
from .exceptions import *
from .expression import Expr
from .description import *
from .unimplemented import UnImplemented, Unknown
from .utilsextension import (
    blosc_compcode_to_compname_ as blosc_compcode_to_compname,
)
from .utilsextension import (
    blosc2_compcode_to_compname_ as blosc2_compcode_to_compname,
)
from .utilsextension import blosc_get_complib_info_ as blosc_get_complib_info
from .utilsextension import blosc2_get_complib_info_ as blosc2_get_complib_info
from .utilsextension import (
    blosc_compressor_list,
    blosc2_compressor_list,
    is_hdf5_file,
    is_pytables_file,
    which_lib_version,
    set_blosc_max_threads,
    set_blosc2_max_threads,
    silence_hdf5_messages,
)

# List here only the objects we want to be publicly available
__all__ = [
    # Exceptions and warnings:
    "HDF5ExtError",
    "ClosedNodeError",
    "ClosedFileError",
    "FileModeError",
    "NaturalNameWarning",
    "NodeError",
    "NoSuchNodeError",
    "UndoRedoError",
    "UndoRedoWarning",
    "PerformanceWarning",
    "FlavorError",
    "FlavorWarning",
    "FiltersWarning",
    "DataTypeWarning",
    "ChunkError",
    "NotChunkedError",
    "NotChunkAlignedError",
    "NoSuchChunkError",
    # Functions:
    "is_hdf5_file",
    "is_pytables_file",
    "which_lib_version",
    "copy_file",
    "open_file",
    "print_versions",
    "test",
    "split_type",
    "restrict_flavors",
    "set_blosc_max_threads",
    "set_blosc2_max_threads",
    "silence_hdf5_messages",
    # Helper classes:
    "IsDescription",
    "Description",
    "Filters",
    "Cols",
    "Column",
    "ChunkInfo",
    # Types:
    "Enum",
    # Atom types:
    "Atom",
    "StringAtom",
    "BoolAtom",
    "IntAtom",
    "UIntAtom",
    "Int8Atom",
    "UInt8Atom",
    "Int16Atom",
    "UInt16Atom",
    "Int32Atom",
    "UInt32Atom",
    "Int64Atom",
    "UInt64Atom",
    "FloatAtom",
    "Float32Atom",
    "Float64Atom",
    "ComplexAtom",
    "Complex32Atom",
    "Complex64Atom",
    "Complex128Atom",
    "TimeAtom",
    "Time32Atom",
    "Time64Atom",
    "EnumAtom",
    "PseudoAtom",
    "ObjectAtom",
    "VLStringAtom",
    "VLUnicodeAtom",
    # Column types:
    "Col",
    "StringCol",
    "BoolCol",
    "IntCol",
    "UIntCol",
    "Int8Col",
    "UInt8Col",
    "Int16Col",
    "UInt16Col",
    "Int32Col",
    "UInt32Col",
    "Int64Col",
    "UInt64Col",
    "FloatCol",
    "Float32Col",
    "Float64Col",
    "ComplexCol",
    "Complex32Col",
    "Complex64Col",
    "Complex128Col",
    "TimeCol",
    "Time32Col",
    "Time64Col",
    "EnumCol",
    # Node classes:
    "Node",
    "Group",
    "Leaf",
    "Table",
    "Array",
    "CArray",
    "EArray",
    "VLArray",
    "UnImplemented",
    "Unknown",
    # The File class:
    "File",
    # Expr class
    "Expr",
]

if "Float16Atom" in locals():
    # float16 is new in numpy 1.6.0
    __all__.extend(("Float16Atom", "Float16Col"))

if "Float96Atom" in locals():
    __all__.extend(("Float96Atom", "Float96Col"))
    __all__.extend(("Complex192Atom", "Complex192Col"))  # XXX check

if "Float128Atom" in locals():
    __all__.extend(("Float128Atom", "Float128Col"))
    __all__.extend(("Complex256Atom", "Complex256Col"))  # XXX check


def get_pytables_version() -> str:
    warnings.warn(
        "the 'get_pytables_version()' function is deprecated and could be "
        "removed in future versions. Please use 'tables.__version__'",
        DeprecationWarning,
    )
    return __version__


def get_hdf5_version() -> str:
    warnings.warn(
        "the 'get_hdf5_version()' function is deprecated and could be "
        "removed in future versions. Please use 'tables.hdf5_version'",
        DeprecationWarning,
    )
    return hdf5_version
